/* 	
	arp.c 	part of TRIPP Traffic Rewriter for IP (v4) Packets
	by poplix@papuasia.org 2005	

*/

#include "tripp.h"

extern int debug;

#define ARP_DEBUG(d,x,...)if(d)print_debug("arp:" ,x,## __VA_ARGS__)
#define ARP_WARNING(w,x,...)if(w)print_warning("arp: " ,x,## __VA_ARGS__)
#define ARP_ERROR(x,...)print_error(0,"arp error: " ,x,## __VA_ARGS__)
//come si notera dai commenti seguenti questo arp funziona solo con ethernet (come tutto il resto)

extern pthread_mutex_t libnet_mutex;
extern pthread_mutex_t malloc_mutex;



TAILQ_HEAD(tqhead, arpentry) arptq;

struct arpentry *arp_allocentry(u_int32_t , char *);

void arpinit(){
	TAILQ_INIT(&arptq);
}

int arpadd(u_int32_t ip, char* mac){
	struct arpentry *fp;
	struct tqhead *tqp;

	tqp = &arptq;
	
	fp=arp_allocentry(ip,mac);
	TAILQ_INSERT_HEAD(tqp,fp,next);
	
	return 1;
}


char *arpget(u_int32_t ip){
	struct arpentry *fp;
	struct tqhead *tqp;
	static char macdst[6];
	tqp = &arptq;	
	
	for (fp = TAILQ_FIRST(tqp); fp != NULL; fp = TAILQ_NEXT(fp, next))
		if(fp->ip == ip){
			memcpy(macdst,fp->mac,6);
			return macdst;
	
		}
		
	return NULL;
	
}


int arpprint(){//unused
	struct arpentry *fp;
	struct tqhead *tqp;

	tqp = &arptq;	
	
	for (fp = TAILQ_FIRST(tqp); fp != NULL; fp = TAILQ_NEXT(fp, next))
		printf("%s\n",int_to_ip(fp->ip));
		
	return 0;
	
}


struct arpentry *
arp_allocentry(u_int32_t ip, char* mac)
{
	struct arpentry *tmp;

	tmp = (struct arpentry *)malloc(sizeof(struct arpentry));
	if (tmp == NULL)
		err(1, "malloc");

	tmp->ip = ip	;
	memcpy(tmp->mac,mac,6);
	
	return(tmp);
}




void arp_pcap_callback(u_char *param, const struct pcap_pkthdr *header, const u_char *pkt_data){
	struct arpframe *af;
	//14 e' ethernet ..funz solo con eth..
	af=(struct arpframe*)(pkt_data +14 +6); //+6 e' la famosa parte iniziale mancante nella struct
		if(ntohs(af->ar_type)==0x02)
			memcpy(param,af,sizeof(struct arpframe));
		else 
			memset(param,0,sizeof(struct arpframe));
		
		
		return;
		
}






int arp_request(u_int32_t ipaddr,libnet_t* l,struct tunnel *tun){
	char ethbcast[6]={0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
	
	//printf("-.->arp req\n");
	libnet_build_arp(
            ARPHRD_ETHER,                           /* hardware addr */
            ETHERTYPE_IP,                           /* protocol addr */
            6,                                      /* hardware addr size */
            4,                                      /* protocol addr size */
            ARPOP_REQUEST,
            tun->outif.l2addr,
            (u_int8_t*)&(tun->outif.ipaddr),
            ethbcast, // dont care... !! 
			(u_int8_t*)&ipaddr,
			NULL,
			0,
			l,
			0);
			
	if((libnet_autobuild_ethernet(ethbcast,ETHERTYPE_ARP,l))==-1)
		ARP_ERROR("libnet error: %s\n", libnet_geterror(l));
 	//writeout()...           
 	if((libnet_write(l))==-1)
 		ARP_ERROR("libnet error: %s\n", libnet_geterror(l));
 	libnet_clear_packet(l);
 	return 1;
 }





/////////////////////////////////////////////////////////////////
char* arp_resolve(u_int32_t ipaddr,libnet_t* l,struct tunnel *tun){
	
 int n=1;	
 pcap_t *pcap_hnd;
 char p_errbuf[PCAP_ERRBUF_SIZE];
 	
 struct bpf_program filter;
 bpf_u_int32 mask;
 bpf_u_int32 net;
 	
 static u_char mac[6];
 struct arpframe arpf;
 int cnt=0;
 
 
	pcap_lookupnet(tun->outif.name, &net, &mask, p_errbuf);
  	pcap_hnd=pcap_open_live(tun->outif.name, DEFAULT_SNAPLEN, 1, 0, p_errbuf);
  	
  	#ifdef BSD
  	  ioctl(pcap_fileno(pcap_hnd), BIOCIMMEDIATE, &n);
    #endif
    
    if(pcap_hnd==NULL)
		ARP_ERROR("pcap Error: %s\n",p_errbuf);
		
  
    pcap_compile(pcap_hnd, &filter, "arp", 0,mask);
    pcap_setfilter(pcap_hnd, &filter);
    pcap_setnonblock(pcap_hnd,1,p_errbuf);
  
    memset(mac,0,6);
    //pthread_mutex_lock(&libnet_mutex);
 	 arp_request(ipaddr,l,tun);
 	//pthread_mutex_unlock(&libnet_mutex);
 	/*
		printf("\nerror writing arp request\n");
		pcap_close(pcap_hnd);
		return NULL;
	}*/
	
    while(1){
		if(cnt==3)break;
	 	pcap_dispatch(pcap_hnd,1,arp_pcap_callback,(u_char*)&arpf);	
		cnt++;
		if(arpf.ar_spa == ipaddr){
			memcpy(mac,arpf.ar_sha,6);
			ARP_DEBUG(debug,"arp resolved %s\n",int_to_ip(ipaddr));
			pcap_close(pcap_hnd);
			return mac;
		}
	usleep(1000000/5);
  	}

   	pcap_close(pcap_hnd);
   	ARP_WARNING(WARNINGS,"host %s not in ethernet",int_to_ip(ipaddr));
   	return NULL;
   
}


