/* 	
	fakearpd.c 	part of TRIPP Traffic Rewriter for IP (v4) Packets
	by poplix@papuasia.org 2005	

*/

#include "tripp.h"

extern int debug;

#define FAKEARPD_DEBUG(d,x,...)if(d)print_debug("fakearpd:" ,x,## __VA_ARGS__)
#define FAKEARPD_WARNING(w,x,...)if(w)print_warning("fakearpd: " ,x,## __VA_ARGS__)
#define FAKEARPD_ERROR(x,...)print_error(0,"fakearpd error: " ,x,## __VA_ARGS__)
//come si notera dai commenti seguenti questo arp funziona solo con ethernet (come tutto il resto)

//libnet handle
libnet_t *libnet_hnd;
//libnet errbuff
char ln_errbuf[LIBNET_ERRBUF_SIZE];

extern pthread_mutex_t libnet_mutex;
extern pthread_mutex_t malloc_mutex;

TAILQ_HEAD(tqhead_fa, fakearpdentry) *fakearpdtq = NULL;

struct fakearpdentry * fakearpd_addentry(u_int32_t ipaddr,u_char *mac){

	struct fakearpdentry *new;
	
	SMALLOC(new,struct fakearpdentry );
	
	new->ip=ipaddr;
	memcpy(new->mac,mac,6);
	
	TAILQ_INSERT_HEAD(fakearpdtq,new,next);
	
	return new;

}

int fakearpd_getentry(u_int32_t ipaddr,u_char *mac){

  struct fakearpdentry *entry;

	for(entry = TAILQ_FIRST(fakearpdtq); entry != NULL; entry = TAILQ_NEXT(entry, next)){
		if(entry->ip == ipaddr){
			memcpy(mac,entry->mac,6);
			
			return 1;
		}
	}
	
	return 0;


}



int fakearpd_entryexists(u_int32_t ipaddr){

  struct fakearpdentry *entry;

	for(entry = TAILQ_FIRST(fakearpdtq); entry != NULL; entry = TAILQ_NEXT(entry, next))
		if(entry->ip == ipaddr)return 1;
	
	return 0;

}





void fakearpd_callback(u_char *param, const struct pcap_pkthdr *header, const u_char *pkt_data){
	struct arphdr_eth *af;
	u_char mac[6];
	int a;
	struct tunnel *tun=(struct tunnel*)param;
	//14 e' ethernet ..funz solo con eth..
	af=(struct arphdr_eth*)(pkt_data +14 ); 
	
	if(*((int*)(af->ar_spa)) == tun->outif.ipaddr) return;
	
	if(ntohs(af->ar_op)==ARPOP_REQUEST)
		if(fakearpd_getentry(*((int*)(af->ar_tpa)),mac)){
			FAKEARPD_DEBUG(debug,"arp request for %s",int_to_ip(*((int*)(af->ar_tpa))));
			arp_reply(*((int*)(af->ar_tpa)),af->ar_sha, tun->outif.ipaddr,tun->outif.l2addr);	
		}
	
	return;
		
}






int arp_reply(u_int32_t ipdst,u_char *dstmac,u_int32_t ipsrc,u_char *srcmac){
  //char ethbcast[6]={0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
	pthread_mutex_lock(&libnet_mutex);
	libnet_build_arp(
            ARPHRD_ETHER,                           /* hardware addr */
            ETHERTYPE_IP,                           /* protocol addr */
            6,                                      /* hardware addr size */
            4,                                      /* protocol addr size */
            ARPOP_REPLY,
            srcmac,
            (u_int8_t*)&ipdst,
            dstmac,
			(u_int8_t*)&ipsrc,
			NULL,
			0,
			libnet_hnd,
			0);
			
	if((libnet_autobuild_ethernet(dstmac,ETHERTYPE_ARP,libnet_hnd))==-1)
		FAKEARPD_ERROR("libnet error: %s\n", libnet_geterror(libnet_hnd));
 	
 	//writeout()...           
 	if((libnet_write(libnet_hnd))==-1)
 		FAKEARPD_ERROR("libnet error: %s\n", libnet_geterror(libnet_hnd));
 	
 	libnet_clear_packet(libnet_hnd);
	pthread_mutex_unlock(&libnet_mutex);
 	return 1;
 }



int init_fakearpdtq(){
	if(fakearpdtq == NULL){
	 	SMALLOC(fakearpdtq,struct tqhead_fa);
 		TAILQ_INIT(fakearpdtq);
 	}
 }




int is_fakearpd_init(){
		if(fakearpdtq == NULL)return 0;
		return 1;
}


char * fakearpd_getfilterstr(){
  static char filter[1024];
  struct fakearpdentry *entry;
  *filter=0;
  
  	filter[0]=0;
  	if(is_fakearpd_init())
	  for(entry = TAILQ_FIRST(fakearpdtq); entry != NULL; entry = TAILQ_NEXT(entry, next))
		sprintf(filter,"%s or dst host %s",filter,int_to_ip(entry->ip));
	  
	return filter;
}


/////////////////////////////////////////////////////////////////




void* fakearpd(void *p){
	
 int n=1;	
 pcap_t *pcap_hnd;
 char p_errbuf[PCAP_ERRBUF_SIZE];
 	
 struct bpf_program filter;
 bpf_u_int32 mask;
 bpf_u_int32 net;
 
 struct tunnel *tun=(struct tunnel *)p;
 
    
    //init libnet
	libnet_hnd = libnet_init(tun->mode,tun->outif.name,ln_errbuf);
	if(libnet_hnd==NULL)
		FAKEARPD_ERROR("libnet init failed: %s\n",ln_errbuf);
 
	pcap_lookupnet(tun->outif.name, &net, &mask, p_errbuf);
  	pcap_hnd=pcap_open_live(tun->outif.name, DEFAULT_SNAPLEN, 1, 0, p_errbuf);
  	
  	#ifdef BSD
  	  ioctl(pcap_fileno(pcap_hnd), BIOCIMMEDIATE, &n);
    #endif
    
    if(pcap_hnd==NULL)
		FAKEARPD_ERROR("pcap Error: %s\n",p_errbuf);
		
  
    pcap_compile(pcap_hnd, &filter, "arp", 0,mask);
    pcap_setfilter(pcap_hnd, &filter);
  //  pcap_setnonblock(pcap_hnd,1,p_errbuf);
   pcap_loop(pcap_hnd, 0, fakearpd_callback, (u_char*)tun);
    
   pcap_close(pcap_hnd);
   	
}

