#include "tripp.h"

extern int debug;

#define FRAG_ERROR(x,...)print_error(0,"IP fragmenter: " ,x,## __VA_ARGS__)
#define FRAG_DEBUG(d,x,...)if(d)print_debug("IP fragmenter: " ,x,## __VA_ARGS__)

extern pthread_mutex_t malloc_mutex;

//from libnet1.0
int in_cksum(u_short *addr, int len){
   int sum;
   int nleft;
   u_short ans;
   u_short *w;
   
   sum = 0;
   ans = 0;
   nleft = len;
   w = addr;
   while (nleft > 1) {
   	sum += *w++;
	nleft -= 2;
	}
	if (nleft == 1) {
		*(u_char *)(&ans) = *(u_char *)w;
		sum += ans;
	}
	return (sum);
}


 

int compute_l4_cksum(struct ip_packet* pkt){

 #define CKSUM_CARRY(x) (x = (x >> 16) + (x & 0xffff), (~(x + (x >> 16)) & 0xffff))
 
  int cksum;
  int len;
  
  switch(pkt->ip->ip_p){
  
    case IPPROTO_TCP:{
      struct tcphdr* tcp=(struct tcphdr*) pkt->layer4;
      
        len=(LIBNET_TCP_H+pkt->layer4_options_size+pkt->payload_size);
       // printf("--->%d %d\n",pkt->layer4_options_size,pkt->payload_size);
  		tcp->th_sum=0;//it's probably already set to zero (by load_ip_packet)
  		cksum=in_cksum((u_short*) &(pkt->ip->ip_src),8);
  		cksum+=htons(IPPROTO_TCP+len);
  		cksum+=in_cksum((u_short*) tcp, len);
  		tcp->th_sum=CKSUM_CARRY(cksum);
  		
  		break;
  	}
  	
  	case IPPROTO_UDP:{
  	  struct udphdr* udp=(struct udphdr*) pkt->layer4;
      
        len=(LIBNET_UDP_H+pkt->payload_size);
  		udp->uh_sum=0;//it's probably already set to zero (by load_ip_packet)
  		cksum=in_cksum((u_short*) &(pkt->ip->ip_src),8);
  		cksum+=htons(IPPROTO_UDP+len);
  		cksum+=in_cksum((u_short*) udp, len);
  		udp->uh_sum=CKSUM_CARRY(cksum);
  		
  		break;
  	}
  	
  }
  
}





void reverse_pktlist(struct tqhead_p *h){
struct tqhead_p *nhead;
struct pktlist *el,*next;

SMALLOC(nhead,struct tqhead_p);
TAILQ_INIT(nhead);

for (el = TAILQ_FIRST(h); el != NULL; el=next){
	next = TAILQ_NEXT(el, next);
	TAILQ_INSERT_HEAD(nhead,el,next);
}


*h=*nhead;
return;
}



void rand_pktlist(struct tqhead_p *h){
  struct tqhead_p *nhead;
  struct pktlist *el,*tel,**arr;
  struct timeval tv;
  int a,sz=0,p[2];
	
	//count list elements  
	for (el = TAILQ_FIRST(h); el != NULL;el=TAILQ_NEXT(el, next))sz++;
	
	//allocate array of pointers
	arr=(struct pktlist**)calloc(sz+1,sizeof(struct pktlist*));
	if( arr == NULL) {
		fprintf(stderr,"calloc out of memory\n");
		exit(0);
	}
	
	//load list into array
	for (el = TAILQ_FIRST(h),a=0; el != NULL;el=TAILQ_NEXT(el, next),a++)arr[a]=el;
	
	//arr[sz]=NULL;
	
	//randomize array
	for(a=0; a < (sz*5) ;a++){
		// get 2 random offsets
		p[0]=(unsigned int)(rand()*tv.tv_usec) % sz ;
		p[1]=(unsigned int)(rand()*tv.tv_usec) % sz ;
		// switch 2 elements
		tel=arr[p[0]];
		arr[p[0]]=arr[p[1]];
		arr[p[1]]=tel;
		
	}
	
	//allocate new list head
	SMALLOC(nhead,struct tqhead_p);
	TAILQ_INIT(nhead);
	
	
	//copy the original list to a new one using randomized array
	for(a=0;a<sz;a++){
		SMALLOC(tel,struct pktlist);
		//memcpy(tel,arr[a],sizeof(struct pktlist));
		tel->p=arr[a]->p;
	
		TAILQ_INSERT_HEAD(nhead,tel,next);
		TAILQ_REMOVE(h,arr[a],next);
		free(arr[a]);
	
	}
	free(arr);
	*h=*nhead;
	
	return;
}




int fragment_packet(struct tqhead_p *pkl_h,struct ip_packet* pkt,u_int mtu,int linktype){

  struct ip_packet *frag;
  struct pktlist *fragcell;
    
  
  u_int	iphlen = LIBNET_IPV4_H + pkt->ip_options_size,

		l4hlen = pkt->headers_size - iphlen, 
  	
  		totsize = pkt->payload_size + l4hlen,
  
  		l2size = get_link_size(linktype),
  		
  		nfb = (mtu-iphlen)/8, off=0, fragcounter=1,ip_frag,
  		
  		frsize;
  
  
  if( (pkt->payload_size+pkt->headers_size) <= mtu){
  	SMALLOC(fragcell,struct pktlist);
  	fragcell->p=pkt;	
  	TAILQ_INSERT_TAIL(pkl_h,fragcell,next);
  	return 1;
  }
  
  FRAG_DEBUG(debug,"fragmenting %d bytes (payload+layer4 headers)",totsize);
  
  //adjust layer4 headers size (needed only if l4options has been rewrited)
  ( ((struct tcphdr*)pkt->layer4)->th_off) = (l4hlen >> 2); 
  
  //non posso lasciare a libnet il compito di computare il cksum di layer4 perche 
  //questo sono pacchetti IP e libnet computerebbe il cksum per un singolo fragment
  //mentre a me serve il cksum di tutti i fragments riassemblati
  
  compute_l4_cksum(pkt);
  
  while(totsize>0){
 
		SMALLOC(frag,struct ip_packet);
		
		//last frag => append all remaining data
		//if( totsize <= mtu){
		if( (totsize+iphlen) <= mtu){
			frsize = totsize;
			ip_frag = 0;
		}
		else {
			frsize = (nfb*8);
			ip_frag = IP_MF;
		}
		
		
		//first fragment => preserve layer4 headers
		if(off==0)
			memcpy(frag->raw, pkt->raw, iphlen + l2size + frsize);
		else
		{
			//copy IP headers
			memcpy(frag->raw, pkt->raw, iphlen + l2size);
			//skip layer4 headers
			memcpy(	frag->raw + l2size + iphlen , pkt->layer4 + (off*8), frsize);
		}			
		
		frag->ip_options_size=pkt->ip_options_size;	
		frag->ip = (struct ip *) (BALIGN + frag->raw + l2size);
		frag->layer4 = (BALIGN + frag->raw + l2size + LIBNET_IPV4_H + frag->ip_options_size);
		frag->payload = frag->layer4;
		frag->headers_size=LIBNET_IPV4_H + frag->ip_options_size; 
		frag->payload_size=frsize;
		  
		frag->counter=1;
		frag->deelay=0; //unused
		
		//serve davvero htons ??? 
		frag->ip->ip_off =  htons(ip_frag | off | (pkt->ip->ip_off & ~IP_OFFMASK) );
		
		SMALLOC(fragcell,struct pktlist);
		fragcell->p=frag;
		  
		  
		TAILQ_INSERT_TAIL(pkl_h,fragcell,next);  
		 
		off+=nfb; 
		totsize-=frsize;
		FRAG_DEBUG(debug,"fragment %d size:%d",fragcounter,frag->payload_size);
		fragcounter++;
  }

  if(pkt->fragorder==FRAGORDER_REVERSE)reverse_pktlist(pkl_h);
  if(pkt->fragorder==FRAGORDER_RANDOM)rand_pktlist(pkl_h);
  
  free(pkt);
  return 1;
}

