/* 	
	ip.c 	part of TRIPP Traffic Rewriter for IP (v4) Packets
	by poplix@papuasia.org 2005

*/

#include "tripp.h"

extern int debug;

#define IP_ERROR(x,...)print_error(0,"IP error: " ,x,## __VA_ARGS__)
#define IP_DEBUG(d,x,...)if(d)print_debug("arp:" ,x,## __VA_ARGS__)

extern pthread_mutex_t malloc_mutex;

u_int ip_to_int(char *ip,int *err){ 

	int a,
		c=0,
		pos=0,
		tmpint;
		
	char t[4];
	u_char  addr[4];
	
	for(a=0; pos<4; a++){	
		if(ip[a]=='.' || ip[a]==0){
		  	t[c]=0;
		  	tmpint=strtol(t, (char **)NULL, 10);
    		if(errno==EINVAL || (tmpint<0 || tmpint>255) )goto bad;
    		addr[pos]=(u_char)tmpint;
			pos++;
			c=0;
		}else {
			t[c]=ip[a];
			c++;
		}
	}
	
	if(pos!=4)goto bad;
	if(err!=NULL)*err=0;
	return *(u_int*)addr;
	
bad:
	if(err!=NULL)*err=1;
	return 0;
}





char *int_to_ip(u_int32_t ip){
	u_char *tmp=(u_char *)&ip;
	static char ret[16];
	sprintf(ret,"%d.%d.%d.%d",tmp[0] & 0xff,tmp[1] & 0xff,tmp[2] & 0xff,tmp[3] & 0xff);
	return ret;
	
}





 
int loadtcpoptions(struct ip_packet *pkt,char *buff,int buffsize){
  
  u_char kind,size;
  struct tcpopt *opts;
  int a,orindex=0;

  //opts=(struct tcpopt*)malloc(sizeof(struct tcpopt));
  
  pkt->decoded_tcp_options.eool=NULL;
  pkt->decoded_tcp_options.noop=NULL;
  pkt->decoded_tcp_options.wss=NULL;
  pkt->decoded_tcp_options.mss=NULL;
  pkt->decoded_tcp_options.ts=NULL;

//  return 1;
  for(a=0;a<buffsize;a++){
  kind=*(buff+a); 
  //pkt->decoded_tcp_options.ord[orindex]=kind;
  //orindex++;
 //if( a >= 100){printf("tho\n");exit(0);}
	switch(kind){
		case TCPOPT_EOOL:
		  pkt->decoded_tcp_options.eool= (buff+a);

		  break;
		case TCPOPT_NOOP:
		  pkt->decoded_tcp_options.noop=(buff+a);
		  break;
		case TCPOPT_SACKP:
		  pkt->decoded_tcp_options.sakp=(u_short*)(buff+a);
		  a+=1;
		  break;
		case TCPOPT_MSS:
		  pkt->decoded_tcp_options.mss=(buff+a);
		  pkt->decoded_tcp_options.mss_size=(*(buff+a+1))-2;
		  a+=pkt->decoded_tcp_options.mss_size+1;
  		  break;
		case TCPOPT_WSS:
		  pkt->decoded_tcp_options.wss=(buff+a);
		  pkt->decoded_tcp_options.wss_size=(*(buff+a+1))-2;
		  a+=pkt->decoded_tcp_options.wss_size+1;
		  break;
		case TCPOPT_TS:
		  //printf("-------------->load TS\n");
		  pkt->decoded_tcp_options.ts=(buff+a);
		  pkt->decoded_tcp_options.ts_size=(*(buff+a+1))-2;
		   //if( *(buff+a+1) != 10){printf("tho2 %d\n",*(buff+a+1));exit(0);}
		  a+=pkt->decoded_tcp_options.ts_size+1;
		  break;
		default:
			IP_DEBUG(debug,"unknown tcp option %d",kind);
			a+=(*(buff+a+1))-1;
			//exit(0);
			//return NULL;
	}
  }
  //printf("oi:%d\n",orindex);

 // pkt->decoded_tcp_options.ord[orindex]=-1;
   
  return 1;
  }
  
  
  
  
  
  




int load_ip_packet(const u_char* pkt_data,struct ip_packet* pkt,u_int rawlen,int linktype){
	int size_link=get_link_size(linktype);
	u_int l4h_off;
	u_int ip_hl;
	//sto usando un buffer statico....pensaci pero..
	//pkt->raw=(u_char*) calloc(rawlen + BALIGN, 1);
	
	//pkt->ip_options=NULL; // dont care
	pkt->ip_options_size=0;
	//pkt->layer4_options=0; // dont care
	pkt->layer4_options_size=0;
	
	//memset(pkt->raw,0,MAXBUFFSIZE);
	memcpy(pkt->raw + BALIGN ,pkt_data,rawlen);
	
	pkt->ip = (struct ip *) (BALIGN + pkt->raw + size_link);
	
	//ip header len
	ip_hl=pkt->ip->ip_hl << 2; //no ntohl
	// decoding ip options
	if(ip_hl > LIBNET_IPV4_H){
		pkt->ip_options=(u_char*) (BALIGN + pkt->raw + size_link + LIBNET_IPV4_H);
		pkt->ip_options_size= ip_hl - LIBNET_IPV4_H;
	}


	//deve essere castato di volta in volta da ogni decoder di livello 4
	pkt->layer4 = (BALIGN + pkt->raw + size_link + LIBNET_IPV4_H + pkt->ip_options_size);
	
	//printf("...>%s\n",int_to_ip(pkt->ip->ip_src.s_addr));
	
	pkt->payload_size=( ntohs(pkt->ip->ip_len) ) -  LIBNET_IPV4_H - pkt->ip_options_size;
	pkt->headers_size=LIBNET_IPV4_H + pkt->ip_options_size; 
	
	pkt->payload=pkt->layer4;
	
	
	//non sono per nulla convinto di ntohs ,, ma cosi funziona,, senza NO ..
	if(	((( ntohs(pkt->ip->ip_off) & ~IP_OFFMASK) & IP_MF ) == 0) &&  	// more_fragments not set
		((  ntohs(pkt->ip->ip_off) & IP_OFFMASK) == 0) )				// ip_off == zero
	
	  switch (pkt->ip->ip_p) {
	
		case IPPROTO_TCP: 
			//malformed packet
			//if( (pkt->payload_size - LIBNET_TCP_H) <= 0 )break; QUESTO MAI ,, MA NON SO XCHE LO HO MESSO...!!
			
			pkt->payload += LIBNET_TCP_H;
			pkt->layer4_options =  pkt->payload;
			l4h_off = ( ((struct tcphdr*)pkt->layer4)->th_off) << 2; //no ntohl
			// decoding tcp options
			if(l4h_off > LIBNET_TCP_H){
			  pkt->layer4_options_size = l4h_off - LIBNET_TCP_H;
			  pkt->payload += pkt->layer4_options_size;
			  loadtcpoptions(pkt,pkt->layer4_options,pkt->layer4_options_size);
			  //printf("%u\n",(*(u_short*)pkt->decoded_tcp_options.wss));
			}
			pkt->payload_size -= (LIBNET_TCP_H + pkt->layer4_options_size);
			pkt->headers_size += (LIBNET_TCP_H + pkt->layer4_options_size);
			//if ipfrag l4 cksum sono computati a mano, altrimenti da libnet
			((struct tcphdr*)pkt->layer4)->th_sum = 0;
			break;
			
 		case IPPROTO_UDP: 
 			 //malformed packet
			//if( (pkt->payload_size - LIBNET_UDP_H) <= 0 )break;
			
 			pkt->payload += LIBNET_UDP_H;
 			//upd has no options
 			pkt->payload_size -= LIBNET_UDP_H;
 			pkt->headers_size += LIBNET_UDP_H;
 			//if ipfrag l4 cksum sono computati a mano, altrimenti da libnet
 			((struct udphdr*)pkt->layer4)->uh_sum = 0;
 			break;
 			
 		case IPPROTO_ICMP:
 		//malformed packet
			//if( (pkt->payload_size - LIBNET_ICMPV4_H) <= 0 )break;
			
 			pkt->payload += LIBNET_ICMPV4_H+4;
 			pkt->payload_size -= LIBNET_ICMPV4_H;//!!!!!!!!!!!!!!!!!!!!!!!
 			pkt->headers_size += LIBNET_ICMPV4_H;
 			//printf("------>ps1:%d\n",pkt->payload_size);
 			//printf("\nicmp %d\n",LIBNET_ICMPV4_H);
 			break;
 	  }
 	  
 	//mai far puntare pk->payload a NULL..infatti mi serve per sapere la lunghezza degli headers
 	
 	//handle (odd) malformed packets 
 	if( pkt->payload_size < 0) pkt->payload_size = 0;
 	
 	pkt->counter=1;
 	pkt->deelay=0;//unused
 	return 1;
}






void writeout(libnet_t* libnet_hnd){//unused
    if(libnet_write(libnet_hnd)==-1)
           	fprintf(stderr, "Write error: %s\n", libnet_geterror(libnet_hnd));

   
   libnet_clear_packet(libnet_hnd);

 }


int build_ip(struct ip_packet *pk,libnet_t *libnet_hnd,u_int mtu){
	libnet_ptag_t lnip;
	if(pk == 0) {
		printf("unexpected null packetA\n");
		exit(9);
	}
	if(pk->payload_size > mtu)IP_ERROR("mtu exceded %d\n",pk->payload_size);
	
	if(pk->payload_size>0)
		libnet_build_data(pk->payload, pk->payload_size, libnet_hnd, 0);	
		//libnet_build_payload(pk->payload, pk->payload_size, libnet_hnd, 0);//libnetNG
		
	if(pk->payload != pk->layer4)//ip packet with layer4
	  switch (pk->ip->ip_p) {
		case IPPROTO_TCP:  
 			build_tcp(pk,libnet_hnd);
 			break;
 		case IPPROTO_UDP:
 			build_udp(pk,libnet_hnd);
  			break;
 		case IPPROTO_ICMP:
 			build_icmp(pk,libnet_hnd);
  			break;
 		//default: return 0;
 		
 		// raw ip
 	  }
 	
 	if(pk->ip_options_size>0)
 		libnet_build_ipv4_options(pk->ip_options,pk->ip_options_size,libnet_hnd,0);
    lnip = libnet_build_ipv4(                                                                          
           pk->headers_size + pk->payload_size,  	     // length 
           pk->ip->ip_tos,                               		// TOS  
           ntohs(pk->ip->ip_id),                 						// IP ID 
           ntohs(pk->ip->ip_off),                                  	// IP Frag 
           pk->ip->ip_ttl,                                 		//TTL 
           pk->ip->ip_p,                        				// protocol 
           0,                                  					// checksum 
           pk->ip->ip_src.s_addr,       						// source IP 
           pk->ip->ip_dst.s_addr,       						// destination IP 
           NULL,                               					// payload 
           0,                                  					// payload size 
           libnet_hnd,                  						// libnet handle 
           0);													// libnet ptag
           
     
     if(lnip==-1){
     	fprintf(stderr, "Eth forging error: %s\n", libnet_geterror(libnet_hnd));
     	return 0;
     }
     return 1;
}





int build_tcp(struct ip_packet *pk,libnet_t *libnet_hnd){
			
	libnet_ptag_t lntcp;
	u_int tcp_size= LIBNET_TCP_H + pk->payload_size  + pk->layer4_options_size;
	struct tcphdr* tcp=(struct tcphdr*) pk->layer4;
	

	if(pk->layer4_options_size>0)
		libnet_build_tcp_options((u_int8_t *)pk->layer4_options, pk->layer4_options_size, libnet_hnd, 0);
	
	
	lntcp = libnet_build_tcp(
        ntohs(tcp->th_sport),            // source port 
        ntohs(tcp->th_dport),            // destination port 
        ntohl(tcp->th_seq),              // sequence number 
        ntohl(tcp->th_ack),              // acknowledgement num 
        tcp->th_flags,                   // control flags 
        ntohs(tcp->th_win),                  	 // window size 
        ntohs(tcp->th_sum),  			                     // checksum 
        ntohs(tcp->th_urp),                     // urgent pointer 
        tcp_size,       // TCP packet size 
	    NULL,                    			 // payload 
        0,                       			 // payload size 
        libnet_hnd,
        0);

    if(lntcp==-1){
    	fprintf(stderr, "tcp forging error: %s\n", libnet_geterror(libnet_hnd));
    	return 0;
    }
    return lntcp;
  }
  
  
  
  

int build_udp(struct ip_packet *pk,libnet_t *libnet_hnd){
			
	libnet_ptag_t lnudp;
	u_int udp_size= LIBNET_UDP_H + pk->payload_size;
	struct udphdr* udp=(struct udphdr*) pk->layer4;
		
       lnudp = libnet_build_udp(
                ntohs(udp->uh_sport),                               /* source port */
                ntohs(udp->uh_dport),                              /* destination port */
                udp_size,           /* packet size */
                ntohs(udp->uh_sum),                                  /* checksum */
                NULL,                            /* payload */
                0,                          /* payload size */
                libnet_hnd,                                  /* libnet handle */
                0); 
                
    if(lnudp==-1){
    	fprintf(stderr, "udp forging error: %s\n", libnet_geterror(libnet_hnd));
    	return 0;
    }
    return lnudp;
  }
  
  
  
  
  
  
  int build_icmp(struct ip_packet *pk,libnet_t *libnet_hnd){
			
	libnet_ptag_t lnicmp;
	u_int icmp_size= LIBNET_ICMPV4_H + pk->payload_size;// + pk->ip_options_size;
	//struct icmphdr* icmp=(struct icmphdr*) pk->layer4;
	struct icmp* icmp=(struct icmp*) pk->layer4;
	//printf("--->%d\n",pk->payload_size);exit(9);
	
	//if(pk->payload_size>0)
		//libnet_build_data(pk->payload, pk->payload_size-4, libnet_hnd, 0);	
	
	//printf("------>ps2:%d\n",pk->payload_size);
		
	switch(icmp->icmp_type){
	  case 0://echo reply
	  	  	
        lnicmp = libnet_build_icmpv4_echo(
            0,                            /* type */
            0,                                    /* code */
            0,                                    /* checksum */
            ntohs(icmp->icmp_hun.ih_idseq.icd_id),                                 /* id */
            ntohs(icmp->icmp_hun.ih_idseq.icd_seq),                                 /* sequence number */
            pk->payload,                                 /* payload */
            pk->payload_size,//!!!!                                    /* payload size */
            libnet_hnd,                                    /* libnet handle */
            0);
	  break;
	  case 8: //echo req
	  	
        lnicmp = libnet_build_icmpv4_echo(
            ICMP_ECHO,                            /* type */
            0,                                    /* code */
            0,                                    /* checksum */
            ntohs(icmp->icmp_hun.ih_idseq.icd_id),                                 /* id */
            ntohs(icmp->icmp_hun.ih_idseq.icd_seq),                                 /* sequence number */
            pk->payload,                                 /* payload */
            pk->payload_size,                                    /* payload size */
            libnet_hnd,                                    /* libnet handle */
            0);
	  break;
	  
	  //default:return -1;
	}
       //lnicmp = 1;
    if(lnicmp==-1){
    	fprintf(stderr, "icmp forging error: %s\n", libnet_geterror(libnet_hnd));
    	return 0;
    }
    return lnicmp;
  }
  
  
  
  