/* 	
	parser.c 	part of TRIPP Traffic Rewriter for IP (v4) Packets
	by poplix@papuasia.org 2005

*/


#include "tripp.h"



//extern TAILQ_HEAD(tqhead_fa, fakearpdentry) *fakearpdtq;

//continue if x char in y string is quotyed
#define CONTINUE_IF_QUOTED(x,y) if( (x>y) && (*(x-1)=='\\'))continue;

#define PARSER_ERROR(x,...)print_error(0,"parser error: " ,x,## __VA_ARGS__)
#define PARSER_WARNING(w,x,...)if(w)print_warning("parser: " ,x,## __VA_ARGS__)

u_char parseflags(u_char **flags_str);
int convertstring(u_char **str,u_char type);
char *nexttoken(char **,char );
u_char getacttype(char *actcmd);

char *pstrchr(char *str,char c);
int removequotes(u_char *buff,u_int *size);

int filter_gettype(char *str,char *type,u_int *size,u_char *l4prot);
//int getvalue(char *str,char *fieldname,char *val_str,int fsize,int *vsize,u_int *opr_not);
int verifyopr(char type,u_char opr, int mos);


//int convert_ascii(char *str,u_int *);

u_char *buildtcpopts(u_char **str,int *optsize);


//elimina tutti i caratteri di tab e gli spazi di tutto quello che non e' tra apici
//esegue inoltre un controllo sugli apici,fondamente perche il parser d'ora in poi fara affidamento
//sulla coerenza degli apici
int check_rulestring(char *str){
    char *t=str;
    for(;*t!=0;t++){
    	CONTINUE_IF_QUOTED(t,str);
    	
		if(*t=='\'')
			//jump after '
			//nota che CONTINUE_IF_QUOTED(t); deve essere chiamato PRIMA di *t=='\'', quindi
			//for(t++;*t!='\'';t++) non funzionerebbe!!
			for(t++;;t++){
				//printf("%c\n",*t);
				CONTINUE_IF_QUOTED(t,str);
				
				if(*t==0)PARSER_ERROR("manca la chiusura di un apice");
					
				if(*t=='\'')break;
			}
	if(*t==' ' || *t=='\t'){
			//il primo par e' la str da trovare e repleace, e siccome t punta a ' ' (o \t)
			//posso passarlo direttamente,
			//vien fuori na roba tipo: trova t e sost con "" in t stesso, perche
			//memrep usa il primo par per trovare l'offset da cui cominciare il replace
			//e l'offs in questo caso e' proprio l'inizio di t (ovvero t stesso)
			memreplace(t,1,"",0,t,strlen(t)+1,1);
			//prima t puntava a " " ma la mem e' stata shiftata indietro di uno => ora t
			//punta al carattere dopo " " quindi il for() mi manda t a due char dopo " "
			t--;
		}

	}
}

int ethalias(char **rulestring,u_char *mac){
	char *ipaddr;
	int converror;
	u_int32_t ip;
	//check_rulestring(*rulestring);

	//trasforma rulestring in tutto quello che viene dopo la prima (
	nexttoken(rulestring, '(');
	//se rulestring e' empty allora vuol dire che nexttoken ha percorso tutta la stringa senza trovare sep
	if(**rulestring==0)PARSER_ERROR("ethalias: missing (");
	
	//trasforma rulestring in tutto quello che viene dopo la prima ) e ritorna un puntare a tutto 
	//quello che era tra graffe

	if(*(*rulestring+(strlen(*rulestring)-1))!=')')
      	PARSER_ERROR("ethalias: missing )");
    
	ipaddr = (char*)nexttoken(rulestring, ')');

	if(**rulestring!=0)PARSER_ERROR("ethalias: error before rule delimiter (';')");
	
	if(!strcmp(ipaddr,*rulestring))PARSER_ERROR("ethalias: missing )");

	ip=ip_to_int(ipaddr,&converror);
	
	if(converror)PARSER_ERROR("ethalias: invalid ip format:%s",ipaddr);
	
	init_fakearpdtq();
	
	fakearpd_addentry(ip,mac); 	
	
}




//legge le regola da un file eliminando tutti gli accapo
int load_rules_table(char *filename,struct tqhead_f* tqpIN,struct tqhead_f* tqpOUT,u_char *mymac){
	FILE *fd;
	char tbuff[1024];
	char fbuff[1024*10];
	char *fp,*bp;
	
	//rulesinit();
	fd=fopen(filename,"r");
	if( fd == NULL)
		PARSER_ERROR("file not found: %s",filename);
		
	while( (fgets(tbuff,sizeof(tbuff),fd))!=NULL){
		if(strlen(tbuff)==0 || tbuff[0]=='#')continue;
		strcat(fbuff,tbuff);
	}
	
	bp=fbuff;
	memreplace("\r",1,"",0,fbuff,strlen(fbuff),-1);
	memreplace("\n",1,"",0,fbuff,strlen(fbuff),-1);
	check_rulestring(bp);
	while((fp=nexttoken(&bp,';'))!=NULL){
	  if(! strncmp(fp,"IN",2) )ruleadd(&fp,tqpIN);
	  
	  else 
	  	if(! strncmp(fp,"OUT",3))ruleadd(&fp,tqpOUT);
	  	else
	  		if(! strncmp(fp,"ethalias",8))ethalias(&fp,mymac);
	  		else PARSER_ERROR("rules must begin with IN or OUT or ethalias ->%s",fp);
	  
	}
	
	fclose(fd);
	return 1;
}






//aggiunge un'azione alla lista delle azioni di una data regola
int actionadd(char type,u_char *rval,u_int rval_size,u_char *mp,u_int mp_size,u_int replace_times,u_char opr,struct tqhead_a *tqp){
	struct actions *fp; 
		
	fp=(struct actions*)malloc(sizeof(struct actions));
	if(fp==NULL)exit(0);
	fp->rtype=type;
	fp->rplc_size=rval_size;
	fp->opr=opr;
	
	if(type==PAYLOAD_REWRITE){
		memcpy(fp->matchpayload,mp,mp_size);
		fp->matchpayload_size=mp_size;
		fp->replacetimes=replace_times;
	}

	memcpy(fp->rplc,rval,rval_size);
	
	TAILQ_INSERT_TAIL(tqp,fp,next);
	return 1;
}






//aggiunge una condizione alla lista delle azioni di una data regola
int matchadd(char type,u_char *mval,u_int mval_size,u_char opr,u_int opr_not,struct tqhead_m *tqp){
	struct matches *fp;
	
	fp=(struct matches*)malloc(sizeof(struct matches));
	if(fp==NULL){printf("malloc error 2\n");exit(0);}
	fp->type=type;
	fp->value_size=mval_size;
	fp->opr=opr;
	fp->opr_not=opr_not;
	
	
	memcpy(fp->value,mval,mval_size);
	
	TAILQ_INSERT_TAIL(tqp,fp,next);
	
	return 1;
}






//aggiunge una regola alla lista
int ruleadd(char **rulestring,struct tqhead_f* tqp){
	struct filter *fp;
	//struct tqhead_f *tqp;
	char 	*ap,
			*matchstr,
			*rplcstr,
			*matchcmd,
			*actcmd;
	
	//head della lista di regola
	//tqp = *tq;
	u_char acttype;
	//check_rulestring(*rulestring);
	
	//trasforma rulestring in tutto quello che viene dopo la prima (
	matchcmd=nexttoken(rulestring, '(');
	//se rulestring e' empty allora vuol dire che nexttoken ha percorso tutta la stringa senza trovare sep
	if(**rulestring==0)goto bad;
	
	//trasforma rulestring in tutto quello che viene dopo la prima ) e ritorna un puntare a tutto 
	//quello che era tra graffe
	ap = (char*)nexttoken(rulestring, ')');
	if(**rulestring==0)goto bad;
	//if(()==NULL)goto bad;
    //printf("%s:%s\n",matchcmd,ap);
    matchstr=ap;
    
    
  //process action string:
       
    actcmd=nexttoken(rulestring, '(');
    if(**rulestring==0)goto bad;
    
    acttype=getacttype(actcmd);
    
    if(acttype==0)PARSER_ERROR("command not found %s\n",actcmd);
	
 	//last char must be ) 
    if(*(*rulestring+(strlen(*rulestring)-1))!=')')goto bad;
      	//PARSER_ERROR("syntax error before rule delimiter (';')\n");
    
    
    //questa chiamata serve solo per sostituire ) con \0 (come *strchr..=0) solo che cosi ignoro la roba tra apici
    ap = (char*)nexttoken(rulestring, ')');
    //anche se l'ultimo char e' ( potrebbe esserci altra roba dopo ; (tipo se uno si dimentica il separatore
    //tra due cmd risultano due cmd attaccati che e' errato pero' l'ultiomo char e' cmq (
    //nota che se l'ultima regola NON finisce con ; il parser non da errore..
    //indi se c'e una sola regola essa puo anche non finire con ;
    if(strcmp(*rulestring,""))
      	PARSER_ERROR("syntax error before rule delimiter (';')\n");
      	
     
    //printf("%s:%s\n\n",actcmd,ap);
	rplcstr=ap;
	
	fp=load_filter(&matchstr,&rplcstr,acttype);
	fp->id=filter_nextindex(tqp);
	
	
	TAILQ_INSERT_TAIL(tqp,fp,next);
  	
  	if(  acttype == ACT_PREWRITE) return 2; //||!!
	return 1;
	
bad:
  PARSER_ERROR("missing ( or )\n");
  
  
}






u_char getacttype(char *actcmd){
	if(!strcmp(actcmd,"set"))return ACT_SET;
	if(!strcmp(actcmd,"clone"))return ACT_CLONE;
	if(!strcmp(actcmd,"drop"))return ACT_DROP;
	if(!strcmp(actcmd,"rewritepayload"))return ACT_PREWRITE;
	if(!strcmp(actcmd,"fragment"))return ACT_FRAGMENT;
	if(!strcmp(actcmd,"print"))return ACT_PRINT;
	if(!strcmp(actcmd,"ask"))return ACT_ASK;
	
	return 0;
}






//parserizza una str nella forma tcp.dport='90' tcp.flags!^'syn+ack'...
int parse_match(char *ap,u_char *mtype,u_char *val,u_int *val_size,u_char *opr,int *opr_not,u_char *l4prot){
  char field[1024];

  
  	u_char *fp;
  	int pt,size;
  	char *p;
	
	u_int val_numericsize;
	fp=val;

	
	if(!strcmp(ap,"tcp")){
		*l4prot=IPPROTO_TCP;
		return FORCE_LAYER4;
	}
	if(!strcmp(ap,"udp")){
		*l4prot=IPPROTO_UDP;
		return FORCE_LAYER4;
	}
	if(!strcmp(ap,"icmp")){
		*l4prot=IPPROTO_ICMP;
		return FORCE_LAYER4;
	}
	memset(field,0,sizeof(field));

	
	if((p=pstrchr(ap,'='))!=NULL)*opr=OPERATOR_EXACT;
	  else if((p=pstrchr(ap,'^'))!=NULL)*opr=OPERATOR_CONTAINS;
	    else if((p=pstrchr(ap,'-'))!=NULL)*opr=OPERATOR_DECREASE;
	      else if((p=pstrchr(ap,'+'))!=NULL)*opr=OPERATOR_INCREASE;
		    else
			  PARSER_ERROR("unknown operator\n");
			  
		    
		
	
	*opr_not=0;
	if(*(p-1)=='!'){
		*opr_not=1;
		p--;
	}
	
	memcpy(field,ap,p-ap);
	
	
	if(!getvalue(p,val,val_size))
		PARSER_ERROR("syntax error (parse_match -> getvalue)\n");
		
	
	
	//converte il fieldname da stringa ad intero (tipo TCP_WIN) per un confronto piu veloce
	val_numericsize=sizeof(field);
	if((pt=filter_gettype(field,mtype,&val_numericsize,l4prot))<1){
		if(pt==0)
			PARSER_ERROR("field not found %s\n",field);
		if(pt==-1)
			PARSER_ERROR("layer4 protocol mismatch\n");
	}
	
	
	if(*mtype!=PAYLOAD)*val_size=val_numericsize;
	
	if(*opr==OPERATOR_CONTAINS && *mtype!=PAYLOAD && *mtype!=TCP_FLAGS)
		PARSER_ERROR("cannot apply '^' operator to %s\n",field);
		
	
	//converte il value da stringa ad intero,a char,,short o da ip(a.b.c.d) ad int
	//e applica htons htonl ecc.. no ecc e basta.. sti due--
	//
	//size is only used for tcp.options
	size=convertstring(&fp,*mtype);
	if(size==-1) PARSER_ERROR("wrong value\n");
	if(*mtype==TCP_OPTIONS)*val_size=size;
	
	if((*mtype== IP_FLAGS_MF || *mtype== IP_FLAGS_MF || *mtype== IP_FLAGS_RF ) &&
			( *(u_char*)fp != 1 && *(u_char*)fp != 0) )
					PARSER_ERROR("ip flags can be set to 1 or 0 only");
	if(*mtype == PAYLOAD)return 10;
	return 1;
}






int parse_prewrite(char **str,u_char *find,u_int *find_size,u_char *replace,u_int *replace_size,int *times){

	char *sap;

		
		if((sap=nexttoken(str,'='))==NULL)
			PARSER_ERROR("missing '='\n");
			
		
		if(!getvalue(sap,find,find_size))
			PARSER_ERROR("syntax error 1\n");
			
		
			
		if((sap=nexttoken(str,'*'))==NULL)
			PARSER_ERROR("missing '*'\n");
			
		
		
		if(!getvalue(sap,replace,replace_size))
			PARSER_ERROR("syntax error 2\n");
			
		
			
		*times=(int)strtol(*str, (char **)NULL, 10);
		if(errno==EINVAL)
			PARSER_ERROR("wrong integer value for 'times of replacements': '%s'",*str);
			
		
		//printf("%s = %s  rs %d\n",find,replace,*replace_size);
					
		
	return 1;
}






//controllo variabili!!!!!

//perserizza la string con la lista delle condizioni e quella con la lista delle azioni
//e per ogniuna crea e popola le due liste all'interno della regola
struct filter *load_filter(char **matchstr,char **rplcstr,u_char actcmd){

  struct filter *f;
  char mtype,rtype;
  
  u_int match_size,rplc_size;
  char *ap;
  u_char 	fval[1024*3],
  			rval[1024*3];
  			
  char 		rfield[1024*3],
  			hfield[1024*3];
  			
  char find[1024*3],replace[1024*6];
  u_int fs,rs;
  int times;
  //questi due pointers servono solo per poter passare un char** a convertstring
  //che potrebbe dover parserizzare una flagsstring (syn+ack...)
  u_char *fp,
  		 *rp;
  			
  u_char l4prot=0,op;
  u_int layer4=0;
  int 	pt,opr_not,val_size,parse;
  
  
  //nuovoa regola
  f=(struct filter *)malloc(sizeof(struct filter));
  if(f == NULL)exit(0);
  
  //inizializza le due liste
  TAILQ_INIT((&(f->actionstq)));
  TAILQ_INIT((&(f->matchestq)));



	
	while((ap = (char*)nexttoken(matchstr, ',')) != NULL){
		//parserizza la str delle condizioni
		match_size=sizeof(fval);
		
		if( (parse_match(ap,&mtype,fval,&match_size,&op,&opr_not,&l4prot)) == FORCE_LAYER4){
			layer4=l4prot;
			continue;
		}
		
		if(!verifyopr(mtype,op,0))
			PARSER_ERROR("operator not allowed in match\n");
		
		//aggiunge alla lista delle condizioni
		if(mtype==PAYLOAD){
			//convert_ascii(fval,&match_size);
			removequotes(fval,&match_size);
		}
		if(l4prot!=0)layer4=l4prot;
		matchadd(mtype,fval,match_size,op,opr_not,(&(f->matchestq)));
	}
	
	
	switch(actcmd){
	  case ACT_SET:
		while((ap = (char*)nexttoken(rplcstr, ',')) != NULL){
			match_size=sizeof(fval);
			
			if( (parse_match(ap,&mtype,fval,&match_size,&op,&opr_not,&l4prot)) == FORCE_LAYER4){
				layer4=l4prot;
				continue;		
			}
			
			if(opr_not)
				PARSER_ERROR("cannot use '!' in set command\n");
				
			
			
			if(!verifyopr(mtype,op,1))
				PARSER_ERROR("operator not allowed in set\n");
			
		
			//aggiunge alla lista delle condizioni
			
			if(mtype==PAYLOAD){
				//convert_ascii(fval,&match_size);
				removequotes(fval,&match_size);
			}
			if(l4prot!=0)layer4=l4prot;
			actionadd(mtype,fval,match_size,NULL,0,0,op,(&(f->actionstq)));
			
		}	
		break;
	  
	  case ACT_CLONE:
	  	f->intval= (int)strtol(*rplcstr, (char **)NULL, 10);
	  	if(errno==EINVAL)
			PARSER_ERROR("clone:wrong value\n");
	  
	  case ACT_FRAGMENT:
	    ap=nexttoken(rplcstr,',');
	    if( ap == NULL)
		    PARSER_ERROR("fragment: missing parameter\n");
	  	f->intval= (int)strtol(ap, (char **)NULL, 10);
	  	if(errno==EINVAL)
			PARSER_ERROR("fragment:wrong fragment size\n");
			
	    ap=nexttoken(rplcstr,',');
	    
	    if( ap == NULL)f->fragorder=FRAGORDER_NORMAL;
		else    
		  if(!strcmp(ap,"random"))f->fragorder=FRAGORDER_RANDOM;
		  else
		    if(!strcmp(ap,"reverse"))f->fragorder=FRAGORDER_REVERSE;
		    else
		      PARSER_ERROR("fragment: fragment order must be reverse or random");

		    
		break;
	  case ACT_PREWRITE:	
		while((ap=nexttoken(rplcstr,','))!=NULL){
			fs=sizeof(find);
			rs=sizeof(replace);
		 	
		 	parse_prewrite(&ap,find,&fs,replace,&rs,&times);
		 	//convert_ascii(replace,&rs);
		 	//convert_ascii(find,&fs);
		 	
		 	removequotes(replace,&rs);
		 	removequotes(find,&fs);
		 	
		 	actionadd(PAYLOAD_REWRITE,replace,rs,find,fs,times,0,(&(f->actionstq)));
		}
		break;
	
	}
	
	f->actcmd=actcmd;
	f->layer4=layer4;
	//printf("FILTER L4:%d\n",layer4);
	//ritorna il pointer alla nuova rule
	return f;
}






//converte il value da strigna a u_char (tipo TCP_WIN) e setta la dimensione in bytes del field per
//il confronto/settaggio
//fa inoltre un confronto sui vari tipi di layer4 per assicurarsi che non vi siano incongruenze tipo
//tcp.win,udp.sport, o match tcp.sport set udp.dport (propagando l4prot)
int filter_gettype(char *str,char *type,u_int *size,u_char *l4prot){

  #define CHECK_LAYER4(x) if(*l4prot==0)*l4prot=x;else if(*l4prot!=x)return -1;

	if(!strcmp(str,"ip.tos")){
		*type=IPTOS;
		*size=1;
		return 1;
	}
	if(!strcmp(str,"ip.id")){
		*type=IP_ID;
		*size=2;
		return 1;
	}
	if(!strcmp(str,"ip.ttl")){
		*type=IPTTL;
		*size=1;
		return 1;
	}
	
	if(!strcmp(str,"ip.protocol")){
		*type=IP_PROTOCOL;
		*size=1;
		return 1;
	}
	
	
	if(!strcmp(str,"ip.src")){
		*type=IP_SRC;
		*size=4;
		return 1;
	}
	if(!strcmp(str,"ip.dst")){
		*type=IP_DST;
		*size=4;
		return 1;
	}
	if(!strcmp(str,"ip.flags.mf")){
		*type=IP_FLAGS_MF;
		*size=1;
		return 1;
	}
	if(!strcmp(str,"ip.flags.df")){
		*type=IP_FLAGS_DF;
		*size=1;
		return 1;
	}
	if(!strcmp(str,"ip.flags.rf")){
		*type=IP_FLAGS_RF;
		*size=1;
		return 1;
	}
	//end of IP
	
	
	if(!strcmp(str,"tcp.sport")){
		CHECK_LAYER4(IPPROTO_TCP);
		*type=TCP_SPORT;
		*size=2;
		return 1;
	}
	if(!strcmp(str,"tcp.dport")){
		CHECK_LAYER4(IPPROTO_TCP);
		*type=TCP_DPORT;
		*size=2;
		return 1;
	}
	if(!strcmp(str,"tcp.seq")){
		CHECK_LAYER4(IPPROTO_TCP);
		*type=TCP_SEQ;
		*size=4;
		return 1;
	}
	if(!strcmp(str,"tcp.ack")){
		CHECK_LAYER4(IPPROTO_TCP);
		*type=TCP_ACK;
		*size=4;
		return 1;
	}
	if(!strcmp(str,"tcp.flags")){
		CHECK_LAYER4(IPPROTO_TCP);
		*type=TCP_FLAGS;
		*size=1;
		return 1;
	}
	if(!strcmp(str,"tcp.win")){
		CHECK_LAYER4(IPPROTO_TCP);
		*type=TCP_WIN;
		*size=2;
		return 1;
	}
	if(!strcmp(str,"tcp.urp")){
		CHECK_LAYER4(IPPROTO_TCP);
		*type=TCP_URP;
		*size=2;
		return 1;
	}
	if(!strcmp(str,"tcp.options")){
		CHECK_LAYER4(IPPROTO_TCP);
		*type=TCP_OPTIONS;
		//*size=2;
		return 1;
	}
	
	if(!strcmp(str,"udp.sport")){
		CHECK_LAYER4(IPPROTO_UDP);
		*type=UDP_SPORT;
		*size=2;
		return 1;
	}
	if(!strcmp(str,"udp.dport")){
		CHECK_LAYER4(IPPROTO_UDP);
		*type=UDP_DPORT;
		*size=2;
		return 1;
	}
	
	


	if(!strcmp(str,"payload")){
		*type=PAYLOAD;
		//!!lascia invariata size!!!!! che chiaramente non si puo sapere a priori
		return 1;
	}
	if(!strcmp(str,"payloadsize")){
		*type=PAYLOADSIZE;
		*size=4;
		return 1;
	}
	return 0;
}




//copia in valstr tutto cio che e' contenuto tra apici
int getvalue(char *str,char *val_str,int *vsize){
	char *p,*end;

	memset(val_str,0,*vsize);
//printf("%s\n",str);
	p=pstrchr(str,'\'');
	if(p==NULL)return 0;
	//jump after the first '\''
	p++;
	//locate the last '\''
	end=pstrchr(p,'\'');
	
	if(end==NULL)return 0;
	
	
	if(*vsize < (end-p)){printf( "buff size not suff\n");return 0;}
//	printf("-->%d\n",*(end+1));
	
	*vsize=(end-p);
	
	memcpy(val_str,p,*vsize);
	//printf("%s\n",val_str);

	if( *(end+1) != 0 )
		PARSER_WARNING(WARNINGS,"garbage after '%s'",val_str);

	return 1;
}






//converte da stringa a intero,short,ipaddr... e memcopia il risultato su buffer puntato da str
// applica htonl... solo se siamo in match, in caso di set hton viene fatto da replace_header
//perche puo essere che il valore debba venir sommato quindi deve arrivare a replace_header in host order
//quella scritta sopra e' una cazzata!
int convertstring(u_char **str,u_char type){
	u_int tmp_int;
	u_short tmp_short;
	u_char tmp_char;
	u_char *tmp_p=*str;
	u_char *tmpbuff;
	u_int error;
	//nota che in caso di indiirzzo ip nella forma a.b.c.d strtoul NON da errore
	if(type!=TCP_FLAGS && type!=PAYLOAD && type!=TCP_OPTIONS){
		tmp_char = tmp_short = tmp_int = (u_int)strtoul(*str, (char **)NULL, 10);
    	if(errno==EINVAL)return(-1);
    }
    
	switch(type){
	  case TCP_FLAGS:
	  	//viene passato tmp_p perche dato che parseflags modifica il parametro se passassi direttamente
	  	//str mi troverei a memcpy non all'inizio del buff puntato da str ma da dove e' finita la stringa 
	  	//che ho parserizzato (per colpa di nexttoken ovviamente)
	  	tmp_char=parseflags(&tmp_p);
	  	memcpy(*str,&tmp_char,1);
	  	break;
	  
	  case TCP_OPTIONS:
	  	//viene passato tmp_p perche dato che parseflags modifica il parametro se passassi direttamente
	  	//str mi troverei a memcpy non all'inizio del buff puntato da str ma da dove e' finita la stringa 
	  	//che ho parserizzato (per colpa di nexttoken ovviamente)
	  	//tmp_char=parseflags(&tmp_p);
	  	tmpbuff=buildtcpopts(&tmp_p,&tmp_int);
	  	memcpy(*str,tmpbuff,tmp_int);
	  	return tmp_int;
	  	break;
	  	
	  //u_char values
	  case IPTOS:
	  case IPTTL:
	  case IP_PROTOCOL:
	  case IP_FLAGS_MF:
	  case IP_FLAGS_DF:
	  case IP_FLAGS_RF:
	  	memcpy(*str,&tmp_char,2);
	  	break;	  	
		
	  //u_short values
	  case IP_ID:
	  case TCP_DPORT:
	  case TCP_SPORT:
	  case TCP_WIN:
	  case TCP_URP:
	  case UDP_SPORT:
	  case UDP_DPORT:
	  	tmp_short=htons(tmp_short);
	  	memcpy(*str,&tmp_short,2);
	  	return tmp_short;
	  	break;
	  
	  //long values
	  case TCP_SEQ:
	  case TCP_ACK:
	  	tmp_int=htonl(tmp_int);
	  	memcpy(*str,&tmp_int,4);
	  	break;
	  
	  //ip addresses
	  case IP_SRC:
	  case IP_DST:
	  	tmp_int=ip_to_int(*str,&error);
	  	if(error)return -1;
	  	//tmp_int=htonl(tmp_int);
	  	memcpy(*str,&tmp_int,4);
	  	break;
	  	
	  case PAYLOAD:
	  	 //do nothing
	  	break;
	  	
	}
	
	return 1;
}





//restituisce un puntatore al primo token 
//e modifica la stringa in ingresso in modo che risulti essere tutto cio che viene dopo il primo separaotre
//nota che tutti i caratteri separatore contenuti fra apici vengono ignorati.
//processare la stringa con: while((ap = (char*)nexttoken(&inputstring, ',')) != NULL)
//	ad ogni ciclo ap sara un token
//nota anche che va in loop infinito se c'e un apice aperto senza uno chiuso
char *nexttoken(char **str,char sep){
	int a;
	char *t=*str,*sw;
	//scorre str usando t come puntatore
	if(*t!=0)
	for(;;t++){

		CONTINUE_IF_QUOTED(t,*str);
		//se trovo un apice sposto il puntatore fino al prossimo apice saltando quelli quotati
		if(*t=='\'')
			for(t++;;t++){
				CONTINUE_IF_QUOTED(t,*str);
				//if( (t>*str) && (*(t-1)=='\\')) continue;
				if(*t == '\'')break;
			}
		//se trovo un sep lo sostituisco con zero,inverto str e t in modo da far puntare str 
		//al primo carattere dopo lo zero e t all'inizio della stringa		
		if(*t == sep){
			*t = 0;
			sw = *str;
			*str = ++t;
			return sw;
		}
		//se la stringa e' finita inverto str e t in modo da far puntare str a zero e t
		//all'inizio della stringa
		if(*t == 0){
			sw = *str;
			*str = t;
			return sw;
		}
	}
	return NULL;
}




//strchr senza considerare la roba tra apici
//non esattamente come strchr perche pstrchr non considera '\0' come parte della string
char *pstrchr(char *str,char c){
char *p=str;
	
	for(; *p!=0; p++){
		
		CONTINUE_IF_QUOTED(p,str);
		
		if(*p == '\'')
		  for(;;p++){
		  	//apice non chiuso
			if(*p == 0)return NULL;
			CONTINUE_IF_QUOTED(p,str);			
			if(*p == '\'')break;
		  }
		  //first occurrence of c found
		  if(*p == c)return p;
	}
	return NULL;
}





//parserizza una string tipo 'syn+ack+urg' e la trasforma nel rispettivo otteto
u_char parseflags(u_char **flags_str){
	u_char *ap;
    u_char flags=0;
         while((ap = (char*)nexttoken((char**)flags_str, '+')) != NULL){
  			if(!strcmp(ap,"syn")){
  				flags |= TH_SYN;	
  				continue;
  			}
  			if(!strcmp(ap,"fin")){
  				flags |= TH_FIN;
  				continue;
  			}
  			if(!strcmp(ap,"rst")){
  				flags |= TH_RST;
  				continue;
  			}
  			if(!strcmp(ap,"push")){
  				flags |= TH_PUSH;
  				continue;
  			}
  			if(!strcmp(ap,"ack")){
  				flags |= TH_ACK;
  				continue;
  			}
  			if(!strcmp(ap,"urg")){
  				flags |= TH_URG;
  				continue;
  			}
  
         	PARSER_ERROR("unknown flag name: %s\n",ap);
         	
         }
    
    return flags;
}





 //remove quotes ('\') e trasforma gli specialchars (\r \n)
int removequotes(u_char *buff,u_int *size){
	u_char *p=buff;
	char err[3];
	char tmp[3];
	int ichr;
  	u_char chr;
  	
	memset(err,0,sizeof(err));
	memset(tmp,0,sizeof(tmp));
	
	for(;(p-buff)<*size; p++){
		CONTINUE_IF_QUOTED(p,buff);
		if( *p=='\\' && ((p+1)-buff) < *size ){
			switch(*(p+1)){
				case 'n':
					*(p+1)='\n';
					break;
				case 'r':
					*(p+1)='\r';
					break;
				case 't':
					*(p+1)='\t';
					break;
				case 'v':
					*(p+1)='\v';
					break;
					
				case ' ':
				case '\t':
					memreplace((p+1),1,"",0,p,*size,1);
					//dato che lavoro su (p+1) anche se la mem viene sciftata
					//NON devo decrementare p
					//p--;
					(*size)--;
					break;
					
				case '\\':
					//faccio si che il memrepl finale rimuova p+1 e lasci indietro p
					// senno al prossimo ciclo viene rimossa p+1 (che e' \ )..
					p++;
					break;
				case '\'':

					break;
				case 'x':

					if( ((p+3)-buff) >= *size )
						PARSER_ERROR("wrong ascii char format, '\\x' char must be followed by two hex digits\n");
						
					memcpy(tmp,(p+2),2);
					ichr=(int)strtol(tmp, (char **)NULL, 16);
					
					//nota che strtol interpreta 7w come 7, ma per me 7w e' errore
					//quindi devo controllare che il secondo digit sia un numero, o una lettera
					//(maiusc o min) da 'a' ad 'f'
					if(errno==EINVAL || ichr < 0 || ichr > 255 ||
					  !(('0' <= tmp[1] && tmp[1] <= '9')  || 
						('a' <= tmp[1] && tmp[1] <= 'f')  || 
						('A' <= tmp[1] && tmp[1] <= 'F') )) 
							 PARSER_ERROR("'\\x%s' is not a valid ascii character\n",tmp);
						
					chr=ichr;
					//sprintf(rpl,"&%s",tmp);
					memreplace(p+1,3,&chr,1,p,*size,1);
					p--;//nota che t-- potrebbe essere suff.
					(*size)-=2;
					break;
					
					
				default:
					memcpy(err,p,2);
					PARSER_ERROR("unknown escape sequence %s",err);
			}
			
			memreplace("\\",1,"",0,p,*size,1);
			p--;
			(*size)--;
		}		
	
	}

}


int filter_nextindex(struct tqhead_f* tqp){
	struct filter *fpf;	
	struct filter *fpl;
	
	
	fpf = TAILQ_FIRST(tqp);
	
	if(fpf == NULL)return 0;
	
	fpl = TAILQ_LAST(tqp,tqhead_f);
	
	if(fpl->id > fpf->id)
	   return fpl->id + 1;
	else 
	   return fpf->id + 1;
		
	
}





//mos is matchorset,..if 0 ->match  if 1 ->set
//non so perche type sia char e non u_char..
int verifyopr(char type,u_char opr, int mos){

	switch(type){
       case IPTOS:
       case IP_ID:
       case IPTTL:
       case TCP_SPORT:
       case	TCP_DPORT:
       case TCP_SEQ:
       case TCP_ACK:
       case TCP_WIN:
       case TCP_URP:
       case UDP_SPORT:
       case UDP_DPORT:
       case IP_PROTOCOL:
      		if(mos==0 && (opr==OPERATOR_EXACT) )return 1;
      		if(mos==1 && (opr==OPERATOR_EXACT || opr==OPERATOR_INCREASE || opr==OPERATOR_DECREASE) )return 1;
       		return 0;
       case IP_SRC:
       case IP_DST:
       case IP_FLAGS_MF:
       case IP_FLAGS_DF:
       case IP_FLAGS_RF:
      		if(mos==0 && (opr==OPERATOR_EXACT) )return 1;
      		if(mos==1 && (opr==OPERATOR_EXACT) )return 1;
       		return 0;
       case TCP_FLAGS:
       		if(mos==0 && (opr==OPERATOR_EXACT || opr==OPERATOR_CONTAINS) )return 1;
       		if(mos==1 && (opr==OPERATOR_EXACT || opr==OPERATOR_INCREASE || opr==OPERATOR_DECREASE) )
       			return 1;
       		return 0;
       case PAYLOAD:
       		if(mos==0 && (opr==OPERATOR_EXACT || opr==OPERATOR_CONTAINS) )return 1;
       		if(mos==1 && (opr==OPERATOR_EXACT) )return 1;
       		return 0;
       case TCP_OPTIONS:
       		if(mos==0 && (opr==OPERATOR_EXACT || opr==OPERATOR_CONTAINS) )return 1;
       		if(mos==1 && (opr==OPERATOR_EXACT) )return 1;
       		return 0;
       		
     }
     
   return 0;
}




//mantieni il valore originale..
//trucchetto sporco: setto il kind al valore corretto pero' con
//size=0.. poi al replace riparserizzo le opzioni e sostituisco
//con il valore iniziale (di pkt) tutte le opz con size=0.
//se il pacchetto che viene generato dal kernel ha opzioni 
//corrotte con size=0 di loro il processo cmq non influira' 
//(risettera' il valore originale  corrotto )
u_char *buildtcpopts(u_char **str,int *optsize){

	u_char 	*ap,
			*res;
	int a=0,
		b;
	u_int ival;
	u_short sval;
	char tmp[20];
	
	
		res=(u_char*)calloc(40,1);
		memset(res,0,40);

         while((ap = (char*)nexttoken((char**)str, '+')) != NULL){
         	if(!strcmp(ap,"noop")){
       			*(res+a)=TCPOPT_NOOP;  		
       			a++;
       			continue;
         	}
          	if(!strcmp(ap,"eool")){
       			*(res+a)=TCPOPT_EOOL;  		
       			a++;
       			continue;
         	}
         	if(!strcmp(ap,"sackp")){
       			*(res+a)=TCPOPT_SACKP;
       			*(res+a+1)=2;
       			a+=(*(res+a+1));
       			continue;
         	}
         	
         	if(!strncmp(ap,"wss=",4)){
       			*(res+a)=TCPOPT_WSS;
       			if(!strncmp( (ap+4),"WSS",3)){
       				*(res+a+1)=0;
       				a+=3;
       				continue;
       			}
       			*(res+a+1)=3;
       			*(res+a+2)=(u_char)strtol((ap+4), (char **)NULL, 10);
       			if(errno==EINVAL)
						PARSER_ERROR("tcp.options.wss: wrong integer value\n");
       			a+=3;
       			continue;
         	}
         	if(!strncmp(ap,"mss=",4)){
       			*(res+a)=TCPOPT_MSS;       			
       			if(!strncmp( (ap+4),"MSS",3)){
       				*(res+a+1)=0;
       				a+=4;
       				continue;
       			}
       			*(res+a+1)=4;
       			sval=(u_short)strtoul((ap+4), (char **)NULL, 10);
       			if(errno==EINVAL)
						PARSER_ERROR("tcp.options.mss: wrong integer value\n");
				sval=htons(sval);
				memcpy((res+a+2),&sval,2);
				a+=4;
       			continue;
         	}
           	if(!strncmp(ap,"ts=",3)){
       			*(res+a)=TCPOPT_TS;       			
       			if(!strncmp( (ap+3),"TS",2)){
       				*(res+a+1)=0;
       				a+=10;
       				continue;
       			}
       			*(res+a+1)=10;
       			for(b=0;  ;b++){
       				
					if( *(ap+3+b) == ','){
					  *(tmp+b)=0;
					  ival=(u_int)strtoul(tmp, (char **)NULL, 10);
					  if(errno==EINVAL)
						PARSER_ERROR("tcp.options.ts[0]: wrong integer value\n");
					  ival=htonl(ival);
					  memcpy((res+a+2),&ival,4);
					  ap+=b+1;
					  b=0;
					}
	
					if( *(ap+3+b) == 0 ){
					  *(tmp+b)=0;
					  ival=(u_int)strtoul((tmp), (char **)NULL, 10);
					  if(errno==EINVAL)
						PARSER_ERROR("tcp.options.ts[1]: wrong integer value\n");
					  ival=htonl(ival);
					  memcpy((res+a+2+4),&ival,4);
					  break;
					 }
					
				*(tmp+b)=*(ap+3+b);
	
       			}	
       			a+=10;
       			continue;
         	}
         	
         	/*
           	if(!strncmp(ap,"WSS",3)){
         		*(res+a)=TCPOPT_WSS;
         		*(res+a+1)=0;
         		a+=3;
         		continue;
         	}
         	if(!strncmp(ap,"MSS",3)){
         		*(res+a)=TCPOPT_MSS;
         		*(res+a+1)=0;
         		a+=4;
         		continue;
         	}
         	
         	
         	if(!strncmp(ap,"TS",2)){

         		*(res+a)=TCPOPT_TS;
         		//memset( (res+a+1) ,0,9);
         		*(res+a+1)=0;
         		a+=10;
         		continue;
         	}
         	*/
         	PARSER_ERROR("unknown tcp option:%s",ap);
         }
*optsize=a;
return res;
}

  
