/* 	
	replacers.c 	part of TRIPP Traffic Rewriter for IP (v4) Packets
	by poplix@papuasia.org 2005

*/

#include "tripp.h"

extern struct tqhead_f *out_rules;
extern struct tqhead_f *in_rules;
extern struct tqhead_c *in_channels;
extern struct tqhead_c *out_channels;

extern pthread_mutex_t malloc_mutex;
extern pthread_mutex_t filter_mutex;
extern int debug;

#define FILTER_ERROR(x,...)print_error(0,"filter error: " ,x,## __VA_ARGS__)
#define FILTER_DEBUG(d,x,...)if(d)print_debug("filter:" ,x,## __VA_ARGS__)


int replace_header(struct ip_packet *pkt, struct tqhead_a *actshead,struct tqhead_f *);
int adjust_tcp_options(u_char *buff,int buffsize,struct ip_packet *pkt);

char *meminmem(char *str,u_int str_size,char *buff,u_int buff_size);

u_char filter_getl4(u_char ft){
	if(0<ft<6)return IPPROTO_IP;
	if(5<ft<13)return IPPROTO_TCP;
	if(12<ft<15)return IPPROTO_UDP;
}

int adjustsynack(int delta,u_int32_t saddr,u_int32_t daddr,u_short sport,u_short dport,int mode);


int print_packet(struct ip_packet*);

//confronta un ip_packet con una lista di condizioni
//nota che tutti i confronti e le assegnazioni vengono fatte a livello di memoria.
//per ogni field da confrontare/settare e' memorizza la dimensione in bytes,dunque qualunque valore (int,short,char)
//viene considerato come stream di bytes
int match_packet(struct ip_packet *pkt,struct tqhead_m *matchh){
	//nota che non posso passare memcmp direttamente a LXOR senno verrebbe richiamata due volte
	#define LXOR(x,y) ( (x || y) && (!(x && y)) )
	
	struct matches *match;

	u_char l4t;
	int cmp;
	//scorre la lista delle condizioni

	for (match = TAILQ_FIRST(matchh); match != NULL; match = TAILQ_NEXT(match, next)){
		l4t=filter_getl4(match->type);
		//se match ip processa altrimenti continue se il paccketto non ha lo stesso l4 della condizione di match
		if(match->type!=PAYLOAD && l4t!=IPPROTO_IP && pkt->ip->ip_p!=l4t)FILTER_ERROR("error1");
		
		
		//in base al tipo di field applica un memcpy tra il val del field e il val della condizione
		switch(match->type){
			case IPTOS:
				cmp=(!memcmp((u_char*)(&(pkt->ip->ip_tos)),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;
					
				FILTER_DEBUG(debug,"pkt_match TOS");
				
				break;
		
			case IP_ID:
				cmp=(!memcmp((u_char*)(&(pkt->ip->ip_id)),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;
				
				FILTER_DEBUG(debug,"pkt_match ID");
			
				break;
			case IPTTL:
				cmp=(!memcmp((u_char*)(&(pkt->ip->ip_ttl)),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;
				FILTER_DEBUG(debug,"pkt_match TTL");
				
				break;
			case IP_PROTOCOL:
				cmp=(!memcmp((u_char*)(&(pkt->ip->ip_p)),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;
				FILTER_DEBUG(debug,"pkt_match IP_PROTOCOL");
				
				break;
				
			case IP_SRC:
				cmp=(!memcmp((u_char*)(&(pkt->ip->ip_src.s_addr)),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;
					
				FILTER_DEBUG(debug,"pkt_match IP SRC");
				
				break;
			case IP_DST:
				cmp=(!memcmp((u_char*)(&(pkt->ip->ip_dst.s_addr)),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;
				FILTER_DEBUG(debug,"pkt_match IP DST");
				break;
				
				
			case IP_FLAGS_MF:
				cmp=( ( pkt->ip->ip_off & IP_MF ));
				if(!LXOR(match->opr_not,cmp))return 0;
				FILTER_DEBUG(debug,"pkt_match IP  FLAGS MF");
				break;	
			case IP_FLAGS_DF:
				cmp=( ( pkt->ip->ip_off & IP_DF ));
				if(!LXOR(match->opr_not,cmp))return 0;
				FILTER_DEBUG(debug,"pkt_match IP FLAGS DF");
				break;
			case IP_FLAGS_RF:
				cmp=( ( pkt->ip->ip_off & IP_RF ));
				if(!LXOR(match->opr_not,cmp))return 0;
				FILTER_DEBUG(debug,"pkt_match IP FLAGS RF");
				break;
				
				
				
			case TCP_SPORT:
				cmp=(!memcmp((u_char*)(&((struct tcphdr*)pkt->layer4)->th_sport),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;
					
				FILTER_DEBUG(debug,"pkt_match TCP SPORT");
				
				break;
				
			case TCP_DPORT:
				cmp=(!memcmp((u_char*)(&((struct tcphdr*)pkt->layer4)->th_dport),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;
					
				FILTER_DEBUG(debug,"pkt_match TCP DPORT");
				
				break;
			case TCP_SEQ:
				cmp=(!memcmp((u_char*)(&((struct tcphdr*)pkt->layer4)->th_seq),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;

				FILTER_DEBUG(debug,"pkt_match SEQ");
				
				break;
			case TCP_ACK:
				cmp=(!memcmp((u_char*)(&((struct tcphdr*)pkt->layer4)->th_ack),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;

				FILTER_DEBUG(debug,"pkt_match ACK");
				
				break;
			case TCP_FLAGS:
				if(match->opr==OPERATOR_EXACT){
				  cmp=(!memcmp((u_char*)(&((struct tcphdr*)pkt->layer4)->th_flags),match->value,match->value_size));
				  if(!LXOR(match->opr_not,cmp))return 0;
					
			      FILTER_DEBUG(debug,"pkt_match flags EXACT %d\n",(u_int)(match->opr_not) );
				 } 
				
				if(match->opr==OPERATOR_CONTAINS){
				//l' ==*(match->value) serve perche senno se chiedo syn+ack lui mi match tutti i pk che hanoo syn
				// e tutti quelli che hanno ack,, invece io voglio che mi match solo quelli che hanno sia syn che ack
				  cmp=( ( (((struct tcphdr*)pkt->layer4)->th_flags) & *(match->value) )==*(match->value));
				  if(!LXOR(match->opr_not,cmp))return 0;
					
					FILTER_DEBUG(debug,"pkt_match flags CONTAINS");
				  
				}
				break;
			case TCP_WIN:
				cmp=(!memcmp((u_char*)(&((struct tcphdr*)pkt->layer4)->th_win),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;
					
				FILTER_DEBUG(debug,"pkt_match WIN");
				
				break;
			case TCP_URP:
				cmp=(!memcmp((u_char*)(&((struct tcphdr*)pkt->layer4)->th_urp),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;
					
				FILTER_DEBUG(debug,"pkt_match URP");
				
				break;
				
				
			case TCP_OPTIONS:
				cmp=(!memcmp(pkt->layer4_options,match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;
					
				FILTER_DEBUG(debug,"pkt_match TCP OPTIONS");
				
				break;
				
				
			case UDP_SPORT:
				cmp=(!memcmp((u_char*)(&((struct udphdr*)pkt->layer4)->uh_sport),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;

				FILTER_DEBUG(debug,"pkt_match UDP SPORT");
				
				break;
			case UDP_DPORT:
				cmp=(!memcmp((u_char*)(&((struct udphdr*)pkt->layer4)->uh_dport),match->value,match->value_size));
				if(!LXOR(match->opr_not,cmp))return 0;
					
				FILTER_DEBUG(debug,"pkt_match UDP DPORT");
				
				break;
			case PAYLOAD:
		  		if(pkt->payload_size==0){
		  			FILTER_DEBUG(debug,"trying to match a packet with empty payload..skipping..");
		  			break;
		  		}
				if(match->opr==OPERATOR_EXACT){
				  //cmp=(!memcmp(pkt->payload,match->value,match->value_size));
				  cmp=(!memcmp(pkt->payload,match->value,pkt->payload_size));
				  if(!LXOR(match->opr_not,cmp))return 0;
					
					
					FILTER_DEBUG(debug,"pkt_match PAYLOAD EXACT %d",(u_int)(match->opr_not) );
				  
				}
				if(match->opr==OPERATOR_CONTAINS){
				  cmp=( (meminmem(match->value,match->value_size,pkt->payload,pkt->payload_size)) != NULL );
				  if(!LXOR(match->opr_not,cmp))return 0;
					
					FILTER_DEBUG(debug,"pkt_match PAYLOAD CONTAINS");
				  
				}
				break;

		  }
		//if(matched==0)return 0;
	}
		
	return 1;
}












struct tcpchan *newtcpchan(u_int32_t saddr,u_int32_t daddr,u_short sport,u_short dport,int val,int next_val,int flag){
	struct tcpchan *tc;
	
	SMALLOC(tc,struct tcpchan);
	
	tc->saddr=saddr;
	tc->daddr=daddr;
	tc->sport=sport;
	tc->dport=dport;
	tc->flag=flag;
	tc->val=val;
	tc->next_val=next_val;
	
	
	return tc;
	
}


struct tcpchan *gettcpchan(u_int32_t saddr,u_int32_t daddr,u_short sport,u_short dport,struct tqhead_c *tqc,int flag){
	struct tcpchan *fp;
	
		for (fp = TAILQ_FIRST(tqc); fp != NULL; fp = TAILQ_NEXT(fp, next))
			if(	fp->daddr==daddr	&&
				fp->saddr==saddr	&&
				fp->sport==sport	&&
				fp->dport==dport	&&
				fp->flag==flag		)
							return fp;


		return NULL;
}




/*
struct tcpchan *gettcpchan(u_int32_t saddr,u_int32_t daddr,u_short sport,u_short dport,struct tqhead_c *tqc,int mode){
	struct tcpchan *fp;
	
		for (fp = TAILQ_FIRST(tqc); fp != NULL; fp = TAILQ_NEXT(fp, next))
		  if(mode == ROLE_IN)
			if(	fp->saddr==saddr	&&
				fp->sport==sport	&&
				fp->dport==dport	)
							return fp;
		  else
		  	
			if(	fp->daddr==daddr	&&
				fp->sport==sport	&&
				fp->dport==dport	)
							return fp;
		return NULL;
}
*/


struct filter *getrule(u_int index,struct tqhead_f* tqp){
	struct filter *fp;
	//struct tqhead_f *tqp;

	//tqp = &filtertq;	
	
	for (fp = TAILQ_FIRST(tqp); fp != NULL; fp = TAILQ_NEXT(fp, next))
		if(fp->id==index)return fp;
		
	return NULL;
	
}



int ask_packet(struct ip_packet* pkt){
  char buff[1024];
  int ask;
	print_packet(pkt);
	
	printf("---> allow this packet?? (y/n): ");
	
	//empty stdin
	//while(getchar() != '\n') continue;
	
	while( (fgets(buff,sizeof(buff),stdin)) != NULL ){
		if(*buff == 'Y' || *buff == 'y'){
		  printf("yes\n");
		  break;
		 }
		if(*buff == 'n' || *buff == 'N'){
			pkt->counter=0;
			break;
		}
		
	}
return 1;
}


//scorre la lista delle regole e le applica all' ip_packet
int process_packet(struct ip_packet *pkt,int mode,int process_all){
	struct filter *fp;
	int psdelta=0;
	
	struct tqhead_f* tqp;
	struct tqhead_c* tqc;
	
	if( mode == ROLE_IN){
		tqp=in_rules;
		tqc=in_channels;
		//printf("\nin\n");
	}
	if( mode == ROLE_OUT){
		tqp=out_rules;
		tqc=out_channels;
		//printf("\nout\n");
	}

	pthread_mutex_lock(&filter_mutex);
	
	for (fp = TAILQ_FIRST(tqp); fp != NULL; fp = TAILQ_NEXT(fp, next)){
	  if(fp->layer4==0 || pkt->ip->ip_p==fp->layer4)
		switch(fp->actcmd){
		  case ACT_PREWRITE:
		  case ACT_SET:
			if(match_packet(pkt, (&(fp->matchestq))) )
				psdelta+=replace_header(pkt, (&(fp->actionstq)),tqp);
			break;
		  case ACT_CLONE:
		  	if( match_packet(pkt, (&(fp->matchestq))) )pkt->counter=fp->intval;
		  case ACT_FRAGMENT:
		  	if( match_packet(pkt, (&(fp->matchestq))) ){
		  		pkt->fragsize=fp->intval;
		  		pkt->fragorder=fp->fragorder;
		  	}
		  	break;
		 case ACT_DROP:
		  	if( match_packet(pkt, (&(fp->matchestq))) )pkt->counter=0;
		  	break;
		  case ACT_PRINT:
		  	if( match_packet(pkt, (&(fp->matchestq))) )print_packet(pkt);
		  	break;
		  case ACT_ASK:
		  	if( match_packet(pkt, (&(fp->matchestq))) )ask_packet(pkt);
		  	break;

		}
	}
	pthread_mutex_unlock(&filter_mutex);
	
	
	
	if(pkt->ip->ip_p == IPPROTO_TCP && psdelta!=0){
		//mantieni la coerenza tra seq e ack una volta modificata la dimensione del payload
		//questa operaz usa un processo parallelo per modificare gli ack e ingannare lo stack locale
		pthread_mutex_lock(&filter_mutex);
		adjustsynack(psdelta,
							pkt->ip->ip_src.s_addr,pkt->ip->ip_dst.s_addr,
							((struct tcphdr*)pkt->layer4)->th_sport,
							((struct tcphdr*)pkt->layer4)->th_dport,
							mode);
		pthread_mutex_unlock(&filter_mutex);
	}
	
	return psdelta;

}


int modifyhfield(u_char* fp,u_char* rplc, u_int size,u_char opr){
	
	u_long lval;
	u_short sval;
	u_char cval;
	
	//u_char values
	if(size==1){
		if(opr==OPERATOR_INCREASE){
		  cval=( (*fp) +  (*rplc) );
		  memcpy((u_char*)fp,(u_char*)&cval,size);
		}
		if(opr==OPERATOR_DECREASE){
		  cval=( (*fp) - (*rplc) );
		  memcpy((u_char*)fp,(u_char*)&cval,size);
		}
		if(opr==OPERATOR_EXACT)
			memcpy((u_char*)fp,(u_char*)rplc,size);
	}
	
	//u_short values
	if(size==2){
		if(opr==OPERATOR_INCREASE){
		  sval=( ntohs( *(u_short*)fp ) + ntohs( *(u_short*)rplc ));
		  sval=htons(sval);
		  memcpy((u_char*)fp,(u_char*)&sval,size);
		}
		if(opr==OPERATOR_DECREASE){
		  sval=( ntohs( *(u_short*)fp ) - ntohs( *(u_short*)rplc ));
		  sval=htons(sval);
		  memcpy((u_char*)fp,(u_char*)&sval,size);
		}
		if(opr==OPERATOR_EXACT)
			memcpy((u_char*)fp,(u_char*)rplc,size);
	}
	
	//u_long values
	if(size==4){
		if(opr==OPERATOR_INCREASE){
		  lval=( ntohl( *(u_int*)fp ) + ntohl( *(u_int*)rplc ));
		  lval=htonl(lval);
		  memcpy((u_char*)fp,(u_char*)&lval,size);
		}
		if(opr==OPERATOR_DECREASE){
		  lval=( ntohl( *(u_int*)fp ) - ntohl( *(u_int*)rplc ));
		  lval=htonl(lval);
		  memcpy((u_char*)fp,(u_char*)&lval,size);
		}
		if(opr==OPERATOR_EXACT)
			memcpy((u_char*)fp,(u_char*)rplc,size);
	}	

return 1;			
}




//simile a match_pachet.. socrre la lista delle azioni ed applica le sostituzioni su ip_packet
//restituisce la differenza di payload_size
int replace_header(struct ip_packet *pkt, struct tqhead_a *actshead,struct tqhead_f *tqp){
	struct actions *act;
	u_char l4t;
	int payload_replacements=0;
	int payload_size_delta=0;
	
	
	//struct tqhead_f *tqp;
	struct filter *f,*fn;
	
	struct tcpchan *tcpc;
	
	u_char opr;
	
	u_short sval;
	u_int 	lval;

	struct actions *modact;

	int newval; //puo essere negativo...
	
	int psdelta=0;
	
	int a;//tmp
	
	u_char tmp_layer4_options[40];
	
	for (act = TAILQ_FIRST(actshead); act != NULL; act = TAILQ_NEXT(act, next)){
		l4t=filter_getl4(act->rtype);
		//se match ip processa altrimenti continue se il paccketto non ha lo stesso l4 della condizione di match
		if(l4t!=0 && pkt->ip->ip_p!=l4t)FILTER_ERROR("error2");
	
		switch(act->rtype){
		  case IPTOS:
			modifyhfield( (u_char*)(&(pkt->ip->ip_tos)), (u_char*)act->rplc,act->rplc_size,act->opr);
			FILTER_DEBUG(debug,"pkt_set TOS");
			break;
		  case IP_ID:
		    modifyhfield( (u_char*)(&(pkt->ip->ip_id)), (u_char*)act->rplc,act->rplc_size,act->opr);
		    FILTER_DEBUG(debug,"pkt_set ID");
			break;
			
			
		  case IP_PROTOCOL:
		    modifyhfield( (u_char*)(&(pkt->ip->ip_p)), (u_char*)act->rplc,act->rplc_size,act->opr);
		    FILTER_DEBUG(debug,"pkt_set PROTOCOL");
			break;
			
		  case IPTTL:
		    modifyhfield( (u_char*)(&(pkt->ip->ip_ttl)), (u_char*)act->rplc,act->rplc_size,act->opr);
		    FILTER_DEBUG(debug,"pkt_set TTL %d",*(u_char*)act->rplc);
			break;
		  case IP_SRC:
		    modifyhfield( (u_char*)(&(pkt->ip->ip_src.s_addr)), (u_char*)act->rplc,act->rplc_size,act->opr);
		    FILTER_DEBUG(debug,"pkt_set IP SRC");
			break;
		  case IP_DST:
		    modifyhfield( (u_char*)(&(pkt->ip->ip_dst.s_addr)), (u_char*)act->rplc,act->rplc_size,act->opr);			
		    FILTER_DEBUG(debug,"pkt_set IP DST");
		    break;
		    

		  case IP_FLAGS_MF:
		    if(*act->rplc == 1)pkt->ip->ip_off |= (IP_MF & IP_OFFMASK) ;
		    if(*act->rplc == 0)pkt->ip->ip_off &=  (IP_MF & IP_OFFMASK);	
		    FILTER_DEBUG(debug,"pkt_set IP_MF");
			break;
		  case IP_FLAGS_DF:
		    if(*act->rplc == 1)pkt->ip->ip_off |= (IP_DF & IP_OFFMASK) ;
		    if(*act->rplc == 0)pkt->ip->ip_off &= (IP_DF & IP_OFFMASK);
		    FILTER_DEBUG(debug,"pkt_set IP_DF");
			break;
		  case IP_FLAGS_RF:
		    if(*act->rplc == 1)pkt->ip->ip_off |= (IP_RF & IP_OFFMASK) ;
		    if(*act->rplc == 0)pkt->ip->ip_off &= (IP_RF & IP_OFFMASK);
		    FILTER_DEBUG(debug,"pkt_set IP_RF");
			break;
			
		  case TCP_SPORT:
			modifyhfield( (u_char*)(&(((struct tcphdr*)pkt->layer4)->th_sport)), (u_char*)act->rplc,act->rplc_size,act->opr);
			FILTER_DEBUG(debug,"pkt_set TCP SPORT");
			break;
		  case TCP_DPORT:
		  	modifyhfield( (u_char*)(&(((struct tcphdr*)pkt->layer4)->th_dport)), (u_char*)act->rplc,act->rplc_size,act->opr);
		  	FILTER_DEBUG(debug,"pkt_set TCP DPORT");
			break;
			
		  case TCP_SEQ:
			modifyhfield( (u_char*)(&(((struct tcphdr*)pkt->layer4)->th_seq)), (u_char*)act->rplc,act->rplc_size,act->opr);
			FILTER_DEBUG(debug,"pkt_set SEQ");
			break;
			
		  case TCP_ACK:
		    modifyhfield( (u_char*)(&(((struct tcphdr*)pkt->layer4)->th_ack)), (u_char*)act->rplc,act->rplc_size,act->opr);
		    FILTER_DEBUG(debug,"pkt_set ACK");
			break;
			
			
		  case TCP_FLAGS:
		  	if(act->opr==OPERATOR_EXACT)
		  		memcpy((u_char*)(&((struct tcphdr*)pkt->layer4)->th_flags),act->rplc,act->rplc_size);
			if(act->opr==OPERATOR_INCREASE)
				((struct tcphdr*)pkt->layer4)->th_flags |= (*((u_char*)act->rplc));
			if(act->opr==OPERATOR_DECREASE) // usa il complementare con ~
				((struct tcphdr*)pkt->layer4)->th_flags &= \
					( (*((u_char*)act->rplc)) ^ ((struct tcphdr*)pkt->layer4)->th_flags );
			
			FILTER_DEBUG(debug,"pkt_set TCP FLAGS");
			break;
			
		  case TCP_WIN:
		    modifyhfield((u_char*)(&((struct tcphdr*)pkt->layer4)->th_win), (u_char*)act->rplc,act->rplc_size,act->opr);
		    FILTER_DEBUG(debug,"pkt_set WIN");
			break;
		  case TCP_URP:
		    modifyhfield((u_char*)(&((struct tcphdr*)pkt->layer4)->th_urp), (u_char*)act->rplc,act->rplc_size,act->opr);
		    FILTER_DEBUG(debug,"pkt_set URP");
			break;
			
		  case UDP_SPORT:
		    modifyhfield((u_char*)(&((struct udphdr*)pkt->layer4)->uh_sport), (u_char*)act->rplc,act->rplc_size,act->opr);
		    FILTER_DEBUG(debug,"pkt_set UDP SPORT");
			break;
		  case UDP_DPORT:
  		    modifyhfield((u_char*)(&((struct udphdr*)pkt->layer4)->uh_dport), (u_char*)act->rplc,act->rplc_size,act->opr);
  		    FILTER_DEBUG(debug,"pkt_set UDP DPORT");
			break;
			
		  case PAYLOAD:
	 		if(pkt->payload_size==0){
		  		FILTER_DEBUG(debug,"try to set payload to a packet with no payload ..skipping..");
		  		break;
		  	}
		  	payload_size_delta=pkt->payload_size;
		  	memcpy(pkt->payload,act->rplc,act->rplc_size);
		  	pkt->payload_size=act->rplc_size;
			payload_size_delta-=pkt->payload_size;
			
			FILTER_DEBUG(debug,"pkt_set PAYLOAD");
			break;
			
		  case PAYLOAD_REWRITE:
		  	if(pkt->payload_size==0){
		  		FILTER_DEBUG(debug,"try to rewrite payload to a packet with no payload ..skipping..");
		  		//break;
		  		return 0; //psdelta=0;()
		  	}
		  	payload_size_delta=pkt->payload_size;
		  	payload_replacements = memreplace(act->matchpayload,act->matchpayload_size,act->rplc,act->rplc_size,pkt->payload,pkt->payload_size,act->replacetimes);
			pkt->payload_size += ( payload_replacements * (act->rplc_size - act->matchpayload_size) );
			payload_size_delta-=pkt->payload_size;

			FILTER_DEBUG(debug,"pkt_REWRITE PAYLOAD");
			break;
			
		  case TCP_OPTIONS:
		  	if( pkt->layer4_options_size != act->rplc_size ){
		  	    //shift payload after the end of new options
				memmove( pkt->layer4_options + act->rplc_size, pkt->payload, pkt->payload_size); 
				pkt->payload += ( act->rplc_size - pkt->layer4_options_size );
				pkt->headers_size += ( act->rplc_size - pkt->layer4_options_size );
				pkt->layer4_options_size=act->rplc_size;
		  	}
		  	memcpy(tmp_layer4_options,act->rplc,act->rplc_size);
		  	adjust_tcp_options(tmp_layer4_options,act->rplc_size,pkt);
		  	memcpy(pkt->layer4_options,tmp_layer4_options,act->rplc_size);
		  	FILTER_DEBUG(debug,"pkt_set TCP_OPTIONS");
		  break;
			
		}
		psdelta+=payload_size_delta;
	}
	
	return psdelta;
}



char *meminmem(char *str,u_int str_size,char *buff,u_int buff_size){
    char *off=NULL;    
    
    int a;
    
    if(buff_size == 0) return NULL;
    
    //find offset per la prima occorrenza di str
   	for(a=0;a<buff_size;a++)
		if(buff[a]==str[0])
			if(!memcmp(str,(buff+a),str_size)){
				off = ( buff + a);
				break;
			}
	
    return off;
 }



//sostituisce una str con un'altra un un buffer
int memreplace(char *str, u_int str_size, char* rplc, u_int rplc_size, char *buff, u_int buff_size,int limit){
    char *off=NULL;

    char *end= (buff+buff_size);
    int len,cnt=0;
    
    if(buff_size == 0) return 0;
    
    //find offset per la prima occorrenza di str

	off=meminmem(str,str_size,buff,buff_size);
    if(off == NULL)return 0;
    
    //fai spazio nel buffer per il replacement (che puo essere piu grande o piu piccolo di str)
    len = end - off;  
    memmove( off + rplc_size, off + str_size, len); 
    memcpy(off,rplc, rplc_size);
	//cnt++;
	//se il buffer non e' finito e il limite de replc imposto non e' raggiunto applica la ricorsione
	limit--;
    if(len+(rplc_size-str_size) >= str_size && limit!=0 )	
		cnt += memreplace(str, str_size, rplc, rplc_size, off + rplc_size,len+((rplc_size-str_size)) ,limit);		
	
	//ritorna il contatore tel tot dei rplc
	return ++cnt;
 }
 
 
 
 

 //problema.. vaietto con tripp su e echo server si inpallina al primo rewrite..
 //sulla porta 80 sembra di no...
 
//applica la differenza dei seq-ack di tutti i pkt di un canare riscritto
//il next_val serve per lasciare intatto il seq o ack al primo giro..
int adjustseqack(u_int32_t saddr,u_int32_t daddr,u_short sport,u_short dport,struct tqhead_c *tqc,struct ip_packet* pkt){
	struct tcpchan *fp;
	struct tcphdr *tcp=(struct tcphdr*)pkt->layer4;
	
	  fp=gettcpchan(saddr,daddr,ntohs(sport),ntohs(dport),tqc,TCP_SEQ);
	  if(fp!=NULL){
	  	  	tcp->th_seq+=htonl(fp->val);
		  //printf("adjustseqack  %d   %d\n",fp->val,fp->flag);
		  fp->val=fp->next_val;
	  }
	  
	  fp=gettcpchan(saddr,daddr,ntohs(sport),ntohs(dport),tqc,TCP_ACK);	  	  	
	  if(fp!=NULL){
	  	  	tcp->th_ack+=htonl(fp->val);
	  	  	//printf("adjustseqack  %d   %d\n",fp->val,fp->flag);
		  fp->val=fp->next_val;
	  }
	 
	  
	
}



//setta il nuovo delta nei tcpchan
int adjustsynack(int delta,u_int32_t saddr,u_int32_t daddr,u_short sport,u_short dport,int mode){
	
	
	struct tcpchan *my_tcpc,*oth_tcpc;
	struct tqhead_c *my_tqc,*oth_tqc;
	
	if(mode == ROLE_IN){
		my_tqc=in_channels;
		oth_tqc=out_channels;

	}
	if(mode == ROLE_OUT){
		my_tqc=out_channels;
		oth_tqc=in_channels;
	}	
	
	
	FILTER_DEBUG(debug,"delta:%d",delta);
	
	my_tcpc=gettcpchan(saddr,daddr,ntohs(sport),ntohs(dport),my_tqc,TCP_SEQ);
	
	//
	//se e' la prima volta che si modifica il payload a questo canale
	//aggiungi la regola e registra il nuovo canale
	if(my_tcpc==NULL){
		
		my_tcpc=newtcpchan(saddr,daddr,ntohs(sport),ntohs(dport),0,delta*(-1),TCP_SEQ);
		TAILQ_INSERT_HEAD(my_tqc,my_tcpc,next);

		//register new channel
		oth_tcpc=newtcpchan(daddr,saddr,ntohs(dport),ntohs(sport),delta,delta,TCP_ACK);
		TAILQ_INSERT_HEAD(oth_tqc,oth_tcpc,next);	

	}
	//altrimenti modifica la regola gia esistente con il nuovo delta
	else {
		my_tcpc->next_val+=(delta*(-1));//!! nextval
		
		oth_tcpc=gettcpchan(daddr,saddr,ntohs(dport),ntohs(sport),oth_tqc,TCP_ACK);
		
		oth_tcpc->next_val+=delta;
		oth_tcpc->val+=delta;
			
	}
				
}





int print_packet(struct ip_packet* pkt){
int sport,dport;

switch(pkt->ip->ip_p){
	case IPPROTO_TCP:{
	  struct tcphdr* tcp=(struct tcphdr*)pkt->layer4;
	  printf("TCP %s.%u > ",int_to_ip(pkt->ip->ip_src.s_addr),ntohs(tcp->th_sport));
	  printf("%s.%u ", int_to_ip(pkt->ip->ip_dst.s_addr),ntohs(tcp->th_dport));
	  	
	  if(tcp->th_flags & TH_SYN)printf("S");
	  if(tcp->th_flags & TH_FIN)printf("F");
	  if(tcp->th_flags & TH_PUSH)printf("P");
	  if(tcp->th_flags & TH_URG)printf("U");
	  if(tcp->th_flags & TH_RST)printf("R");
	  if(tcp->th_flags == 16)printf(".");
	  printf(" %u (%u) ack %u win %u",ntohl(tcp->th_seq),
	  	pkt->payload_size, ntohl(tcp->th_ack),ntohs(tcp->th_win));
	  printf("\n");
	  break;
	}
	
	case IPPROTO_UDP:{
	  struct udphdr* udp=(struct udphdr*)pkt->layer4;
	  printf("UDP %s.%u > ",int_to_ip(pkt->ip->ip_src.s_addr),ntohs(udp->uh_sport));
	  printf("%s.%u ", int_to_ip(pkt->ip->ip_dst.s_addr),ntohs(udp->uh_dport));
	  printf("len: %u\n",udp->uh_ulen);
	  break;
	}
	
	default:
	  printf("UNKNOWN IP PROTOCOL %u %s > ",pkt->ip->ip_p,int_to_ip(pkt->ip->ip_src.s_addr));
	  printf("%s", int_to_ip(pkt->ip->ip_dst.s_addr));
	  printf("\n");
}
//	
}






int adjust_tcp_options(u_char *buff,int buffsize,struct ip_packet *pkt){
  
  u_char kind;
  struct tcpopt *opts;
  int a=0,orindex=0,size;
 
  while(a < buffsize){
   kind=*(buff+a); 
	 switch(kind){
		case TCPOPT_EOOL:
		  size=1;
		  break;
		case TCPOPT_NOOP:
		  size=1;
		  break;
		case TCPOPT_SACKP:
		  size=2;
		  break;
		case TCPOPT_MSS:
		  size=4;
		  if(*(buff+a+1)==0) memcpy( buff+a, pkt->decoded_tcp_options.mss,size);
  		  break;
		case TCPOPT_WSS:
		  size=3;
		  if(*(buff+a+1)==0)memcpy( buff+a, pkt->decoded_tcp_options.wss,size);
		  break;
		case TCPOPT_TS:
		 size=10;
		 if(*(buff+a+1)==0) { 		
		    if (pkt->decoded_tcp_options.ts == NULL) {
			  printf( "anzi no\n");
			  break;
		    }
		    memcpy( buff+a, pkt->decoded_tcp_options.ts,size);
	     }
		  break;
		  
		  default:
		    printf("ADJUST:unknown tcp option %d\n",kind);
			//exit(0);
			size=*(buff+a+1);
	}
	a+=size;
  }

   
  return 1;
  }



