/* 	
	tripp.c 	part of TRIPP Traffic Rewriter for IP (v4) Packets
	by poplix@papuasia.org 2005

     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
     
*/

#include "tripp.h"

#define INIT_ERROR(x,...)print_error(0,"init error: " ,x,## __VA_ARGS__)
#define RT_ERROR(x,...)print_error(0,"runtime error: " ,x,## __VA_ARGS__)
#define TRIPP_DEBUG(d,x,...)if(d)print_debug("" ,x,## __VA_ARGS__)


#define DEVVER "1"

#ifdef BSD
#define DEFAULT_INIF "lo0"
#endif

#ifdef __linux__
#define DEFAULT_INIF "lo"
#endif


struct thparam{
	int singproc;
	struct tunnel *tun;
	//char 	frules[1024];
	struct tqhead_f *filtertq;
	u_int32_t gw;
	
};


pthread_mutex_t filter_mutex;
pthread_mutex_t libnet_mutex;
pthread_mutex_t malloc_mutex;


int debug;

struct tqhead_f *out_rules;
struct tqhead_f *in_rules;
struct tqhead_c *in_channels;
struct tqhead_c *out_channels;


void* tripp_process(void *par);
	
	
int get_link_size(int dl){
	   switch (dl) {    
        case 	DLT_EN10MB:	return sizeof(struct ether_header);
        case  	DLT_LOOP:	return 4;
        default:return -1;
        }
 }
 
 

 
 
 int init_intfs(char *inif_name,char* outif_name,struct tunnel* tun){
	//defaults
	tun->inif.oldmtu=0;
	tun->outif.oldmtu=0;
	
	//getinfo
	if(!getifinfo(outif_name,&(tun->outif)) ||
       !getifinfo(inif_name,&(tun->inif)) ){
  			INIT_ERROR("error fetching interface infos\n");
  			
  	}

	if(tun->outif.ipaddr==0x00)
		INIT_ERROR("no ip for writeif\n");
		
	
	//adjust mtus
	if(tun->inif.mtu > tun->outif.mtu){
		setmtu(&(tun->inif),tun->outif.mtu);
		tun->mtu=tun->outif.mtu;
	}

	if(tun->inif.mtu < tun->outif.mtu){
		setmtu(&(tun->outif),tun->inif.mtu);
		tun->mtu=tun->inif.mtu;
	}
	
	if(tun->inif.mtu == tun->outif.mtu)
		tun->mtu=tun->inif.mtu;
	
	
	return 1;
}
 

void usage(char *appname){
	printf(	"TRIPP ver: %s-%s\n"
			"  usage:\n"
			"    %s -o <interface> [-f <rulesfile>] [-g <gateway>] [-S] [-d]\n\n"
			"      -o      output interface\n"
			"      -f      path to rules file\n"
			"      -g      specify default gateway\n"
			"      -S      start outgoing rewriter only\n"
			"      -d      print debug messages\n\n"
			,VERSION,DEVVER,appname);
	exit(0);
}


//conta i bit a uno in countnm
int countbits(u_int32_t countnm){
  int nmbits;
  
  	for (nmbits = 0; countnm; countnm >>= 1) nmbits += countnm & 1;
  	
  	return nmbits;
}




int main(int argc, char **argv){

	char 	*inintf=NULL,
			*outintf=NULL;
			
	int 	mode=-1,
			singproc=0,
			c;
	
	//parameter for first process
	struct thparam par1;
	//parameter for second process
	struct thparam par2;
	
	//tunnel for first proc
	struct tunnel tun1;
	//tun for second proc
	struct tunnel tun2;
	
	//rulesfile name
    char *filename=NULL;
    
	//gateway ip
    char *gwip=NULL;
    int err=0;
    
    //netaddr in str (ex: 10.0.0.0/24)
    char mynet_str[18+1];
     
    extern char *optarg;
    
	sigset_t sigset;
    int rcvsignal;
    
    pthread_t th_in, th_out, th_ethalias;
    
    
    
	if(argc==1)usage(argv[0]);
	
	debug = 0;
	
     //process commandline options and parameters
     while ((c = getopt(argc, argv, ":dSo:f:g:")) != -1) {
       switch(c) {
         case 'd':
         	debug=1;     
            break;
         case 'S':
         	singproc=1;     
            break;
         case 'g':
            gwip = optarg;  
            break;
         case 'f':
            filename = optarg;  
            break;
         case 'i':
         	inintf = optarg;
			break;
		 case 'o':
			outintf = optarg;
			break;
         case ':':
            printf("error: -%c: argument is mandatory \n", optopt);
            usage(argv[0]);
            break;
         case '?':
            printf("error: unknown option %c\n", optopt);
            usage(argv[0]);
            break;
       }
     }
	
	
	//if(mode==LIBNET_RAW4)inintf=outintf;
	mode=LIBNET_LINK;
	
	//if(inintf==NULL)
	inintf=DEFAULT_INIF;
	
	if(mode==-1 || outintf==NULL)usage(argv[0]);
	
	sigemptyset(&sigset);
	sigaddset(&sigset, SIGINT);
	sigaddset(&sigset, SIGTERM);
	
	pthread_sigmask ( SIG_BLOCK, &sigset, NULL );	

	//set interfaces default values and adjust mtus
	init_intfs(inintf,outintf,&tun1);
 	tun1.mode=LIBNET_LINK;
	par1.tun=&tun1;
	par1.singproc=singproc;
	
		
	tun2.mode=LIBNET_RAW4;
	par2.tun=&tun2;
	
	if(gwip!=NULL){
		par1.gw=ip_to_int(gwip,&err);
		if(err)INIT_ERROR("parsing error:%s",gwip);
	}
	else par1.gw=0;

//fakearpd(&tun1);
	SMALLOC(out_rules,struct tqhead_f);
	SMALLOC(in_rules,struct tqhead_f);
	
	SMALLOC(in_channels,struct tqhead_c);
	SMALLOC(out_channels,struct tqhead_c);
	
	TAILQ_INIT(out_rules);
	TAILQ_INIT(in_rules);
	TAILQ_INIT(out_channels);
	TAILQ_INIT(in_channels);
	
		
	pthread_mutex_init(&filter_mutex,NULL);
	pthread_mutex_init(&libnet_mutex,NULL);
	pthread_mutex_init(&malloc_mutex,NULL);
	
	if(filename != NULL)
 		load_rules_table(filename, in_rules, out_rules,tun1.outif.l2addr);	
	
	
	init_intfs(outintf,outintf,&tun2);

	if(is_fakearpd_init())
		pthread_create(&th_ethalias,NULL,(void*)fakearpd,(void*)&tun1);
	
	sprintf(mynet_str,"%s/%u",
  					int_to_ip(tun1.outif.ipaddr & tun1.outif.netmask),//network address
  					countbits(tun1.outif.netmask)//netmask bits
  					); 
	
	
	
	
	//add routing
  	route(mynet_str,tun1.outif.name,tun1.inif.name);
  	if( gwip != NULL)	
  		route_defgw(gwip,"127.0.0.1",tun1.outif.name,tun1.inif.name);
  	TRIPP_DEBUG(debug,"routing table modifyed");
	
	
	pthread_create (&th_out, NULL,(void *) tripp_process, (void*)&par1 );

	if(!singproc){	
		//sleep(1);
		
		frw_deny(mynet_str,tun2.outif.name);
	  	TRIPP_DEBUG(debug,"firewall rule added");
		
		pthread_create (&th_in, NULL,(void *) tripp_process, (void*)&par2 );		
	
	}
	
	sigwait(&sigset,&rcvsignal);
	
	if(!singproc){
		pthread_cancel(th_in);
		frw_delete(mynet_str,tun2.outif.name);
		TRIPP_DEBUG(debug,"firewall rule deleted");
	}
	
	pthread_cancel(th_out);		
	
	
	if(restoremtu(&(tun1.inif)))TRIPP_DEBUG(debug,"%s MTU restored ",tun1.inif.name);
	if(restoremtu(&(tun1.outif)))TRIPP_DEBUG(debug,"%s MTU restored ",tun1.outif.name);
	
	//delete routing
  	route(mynet_str,tun1.inif.name,tun1.outif.name);
  	if( gwip != NULL)	
  		route_defgw("127.0.0.1",gwip,tun1.inif.name,tun1.outif.name);
	TRIPP_DEBUG(debug,"routing table restored");
	

	//rmsocket();
	printf("Exiting..\n");
	exit(1);
	
}



int rprint(struct tqhead_f* tqp){
	struct filter *fp;
	//struct tqhead_f *tqp;

	//tqp = &filtertq;	
	
	for (fp = TAILQ_FIRST(tqp); fp != NULL; fp = TAILQ_NEXT(fp, next))
		printf("%d\n",fp->id);
		
	return 0;
	
}



void* tripp_process(void *vpar){
	//ioctl parameter
	int n=1;
	
	//pcap handle
	pcap_t *pcap_hnd;
	//pcap errbuff
 	char p_errbuf[PCAP_ERRBUF_SIZE];
 	
 	//pcapp filter
 	struct bpf_program filter;
	bpf_u_int32 mask;
	bpf_u_int32 net;
	//buffer for filterstring
 	u_char filterstring[500];
 	
 	//libnet handle
	libnet_t *libnet_hnd;
	//libnet errbuff
	char ln_errbuf[LIBNET_ERRBUF_SIZE];

	//struct for reading l2addr with libnet
	struct libnet_ether_addr *hwaddr;
	
	//parameter for pcap callbak
	struct pcap_parameter pcp;
	
	//tunnel
	struct tunnel *tun;
	
	//casted parameter for this function
	struct thparam* par=(struct thparam*)vpar;
	
	struct tqhead_f *filtertqp=par->filtertq;
	

	
	tun=par->tun;
		
    //init libnet
	libnet_hnd = libnet_init(tun->mode,tun->outif.name,ln_errbuf);
	if(libnet_hnd==NULL)
		INIT_ERROR("libnet init failed: %s\n",ln_errbuf);
	

  sprintf(filterstring,"(icmp or udp or tcp and net %s/%u)",
  					int_to_ip(tun->outif.ipaddr & tun->outif.netmask),//network address
  					countbits(tun->outif.netmask)//netmask bits
  					); 
 
  

  if(tun->mode == LIBNET_LINK){
    //set outif layer2 address
  	tun->outif.l2addr_size=6;
  	hwaddr=libnet_get_hwaddr(libnet_hnd);
  	memcpy(&(tun->outif).l2addr,&hwaddr->ether_addr_octet,6); 	
 	//init arp table
 	arpinit();
 	sprintf(filterstring,"%s and dst host not %s",filterstring,int_to_ip(tun->outif.ipaddr));
  }else{
  	
  	sprintf(filterstring,"%s and ( dst host %s",filterstring,int_to_ip(tun->outif.ipaddr));
	sprintf(filterstring,"%s %s )",filterstring,fakearpd_getfilterstr());
  	//printf("%s\n",filterstring);
  }
  
  pcap_hnd=pcap_open_live(tun->inif.name, DEFAULT_SNAPLEN, 1, 0, p_errbuf);
 
  if(pcap_hnd==NULL)
	INIT_ERROR("pcap error:%s\n",p_errbuf);
	
  
  #ifdef BSD
   ioctl(pcap_fileno(pcap_hnd), BIOCIMMEDIATE, &n);
  #endif
  
 //compile pcap filter
 if((pcap_compile(pcap_hnd, &filter, filterstring, 0,tun->inif.netmask))==-1)
  	INIT_ERROR("%s %s\n","pcap filter error\n",filterstring);
 //set the filter 	
 pcap_setfilter(pcap_hnd, &filter);
 
 
 if(tun->mode == LIBNET_LINK)
 	printf("rewriter for outgoing traffic started\n");
 else
 	printf("rewriter for incoming traffic started\n");
 
 //pcap parameter
 pcp.libnet_h=libnet_hnd;
 pcp.tun=par->tun;
 pcp.tqp=&par->filtertq;
 pcp.gw=par->gw;
 
 //start sniffing
 pcap_loop(pcap_hnd, 0, pcap_callback, (u_char*)&pcp);

//never reached..
 pcap_close(pcap_hnd);
  
 //return 1;
  
}

//frees pakcet list cell
void free_pklcell(struct pktlist* pl){
    free(pl->p);
    free(pl);
 }
 
 
void pcap_callback(u_char *param, const struct pcap_pkthdr *header, const u_char *pkt_data){
	//casted parameter for this function
	struct pcap_parameter *pcp=(struct pcap_parameter *) param;

	//new ip packet
	struct ip_packet *newpkt;
	
	//pinter to read packes list
	struct pktlist *pkt;
	struct tqhead_p *tqpk;
	
	//clone{n} counter
	int a,psdelta;



	SMALLOC(newpkt,struct ip_packet);
	if( (load_ip_packet(pkt_data,newpkt,header->caplen,pcp->tun->inif.type)) == 0 ){ 
    	free(newpkt);
    	TRIPP_DEBUG(debug,"packet ignored");
		return;
	}
	
	newpkt->fragsize=pcp->tun->mtu;

 	
 	
 	psdelta=process_packet(newpkt,pcp->tun->mode);

 	if(pcp->tun->mode == ROLE_IN )		
	   adjustseqack(
						newpkt->ip->ip_src.s_addr,newpkt->ip->ip_dst.s_addr,
						((struct tcphdr*)newpkt->layer4)->th_sport,
						((struct tcphdr*)newpkt->layer4)->th_dport,
						in_channels,newpkt);
 	
 	if(pcp->tun->mode == ROLE_OUT )		
	   adjustseqack(
						newpkt->ip->ip_src.s_addr,newpkt->ip->ip_dst.s_addr,
						((struct tcphdr*)newpkt->layer4)->th_sport,
						((struct tcphdr*)newpkt->layer4)->th_dport,
						out_channels,newpkt);
 	
 	//incoming  rewriter must decide if newpkt needs to be sent to local tcp stack
 	//if(pcp->tun->mode == ROLE_IN )
 	//	if(newpkt->ip->ip_dst.s_addr != pcp->tun->outif.ipaddr) { free(newpkt);return;}

 	
 	///
 	

 	pthread_mutex_lock(&libnet_mutex);
 	SMALLOC(tqpk,struct tqhead_p);
 	TAILQ_INIT(tqpk);
 	
 	fragment_packet(tqpk,newpkt,newpkt->fragsize,pcp->tun->inif.type);
 	
 	for (pkt = TAILQ_FIRST(tqpk); pkt != NULL; pkt = TAILQ_NEXT(pkt, next)){
 	
		if(pkt->p->counter <= 0) goto close;
		
		if(!build_ip(pkt->p,pcp->libnet_h,pcp->tun->mtu)){
			RT_ERROR("UNEXPECTED ERROR 2\n"); 
			goto close;
		}
	
		if(pcp->tun->mode == LIBNET_LINK)
			if(!build_layer2(pkt->p,pcp->libnet_h,pcp->tun,pcp->gw)) goto close;
		
		for(a=0; a < pkt->p->counter; a++){
			if(libnet_write(pcp->libnet_h)==-1)
				RT_ERROR("Write error: %s\n", libnet_geterror(pcp->libnet_h));
		 }	
		
	close:
		libnet_clear_packet(pcp->libnet_h);
	
	
		free_pklcell(pkt);
		TAILQ_REMOVE(tqpk,pkt,next);		
		

    }
    free(tqpk);
pthread_mutex_unlock(&libnet_mutex);
}
















